import logging
from collections.abc import Callable
from datetime import datetime
from enum import auto
from functools import partial
from pprint import pformat
from typing import TYPE_CHECKING, Any

from mstrio import config
from mstrio.api import documents, reports, subscriptions
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.distribution_services.subscription.content import Content
from mstrio.distribution_services.subscription.delivery import (
    CacheType,
    ClientType,
    Delivery,
    LegacyCacheType,
    LibraryCacheTypes,
    Orientation,
    SendContentAs,
    ShortcutCacheFormat,
    ZipSettings,
)
from mstrio.distribution_services.subscription.subscription_status import (
    SubscriptionStatus,
)
from mstrio.helpers import NotSupportedError
from mstrio.modeling import Prompt
from mstrio.users_and_groups import User
from mstrio.utils import helper, time_helper
from mstrio.utils.entity import EntityBase
from mstrio.utils.enum_helper import AutoUpperName
from mstrio.utils.helper import (
    get_args_from_func,
    get_default_args_from_func,
    get_response_json,
)
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import subscriptions as subscriptions_processors
from mstrio.utils.version_helper import is_server_min_version, method_version_handler

if TYPE_CHECKING:
    from mstrio.server.project import Project


logger = logging.getLogger(__name__)


class RecipientsTypes(AutoUpperName):
    CONTACT_GROUP = auto()
    USER_GROUP = auto()
    CONTACT = auto()
    USER = auto()
    PERSONAL_ADDRESS = auto()
    UNSUPPORTED = auto()


class Subscription(EntityBase):
    """Class representation of Strategy One Subscription object.

    Attributes:
        subscription_id: The ID of the Subscription
        connection: The Strategy One connection object
        project_id: The ID of the project the Subscription belongs to
    """

    _API_GETTERS = {
        (
            "id",
            "name",
            "multiple_contents",
            "editable",
            "date_created",
            "date_modified",
            "owner",
            "schedules",
            "contents",
            "recipients",
            "delivery",
        ): subscriptions.get_subscription,
        "status": subscriptions_processors.get_subscription_status,
        "last_run": subscriptions_processors.get_subscription_last_run,
    }

    @staticmethod
    def _parse_status(source: dict, connection: 'Connection'):
        """Parse Subscription Status from the API response."""

        return SubscriptionStatus.from_dict(source, connection) if source else None

    _FROM_DICT_MAP = {
        "owner": User.from_dict,
        "contents": (
            lambda source, connection: [
                Content.from_dict(content, connection) for content in source
            ]
        ),
        "delivery": Delivery.from_dict,
        "schedules": (
            lambda source, connection: [
                Schedule.from_dict(content, connection) for content in source
            ]
        ),
        "date_created": time_helper.DatetimeFormats.YMDHMS,
        "date_modified": time_helper.DatetimeFormats.YMDHMS,
        "last_run": time_helper.DatetimeFormats.YMDHMS,
        "status": _parse_status,
    }
    _API_PATCH = [subscriptions.update_subscription]
    _RECIPIENTS_TYPES = RecipientsTypes
    _RECIPIENTS_INCLUDE = ['TO', 'CC', 'BCC', None]
    _API_GET_PROMPTS = staticmethod(subscriptions.get_subscription_prompts)

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        subscription_id: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initialize Subscription object, populates it with I-Server data.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            id (str, optional): ID of the subscription to be initialized, only
                id or subscription_id have to be provided at once, if both
                are provided `id` will take precedence
            subscription_id (str, optional): ID of the subscription to be
                initialized
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """

        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
        )

        subscription_id = id or subscription_id
        if not subscription_id:
            helper.exception_handler(
                msg='Must specify valid id or subscription_id',
                exception_type=ValueError,
            )

        super().__init__(connection, subscription_id, project_id=proj_id)

    def _init_variables(self, project_id, **kwargs):
        super()._init_variables(**kwargs)
        self.subscription_id = kwargs.get('id')
        self.multiple_contents = kwargs.get('multiple_contents')
        self.editable = kwargs.get('editable')
        self.allow_delivery_changes = kwargs.get('allow_delivery_changes')
        self.allow_personalization_changes = kwargs.get('allow_personalization_changes')
        self.allow_unsubscribe = kwargs.get('allow_unsubscribe')
        self.date_created = time_helper.map_str_to_datetime(
            "date_created", kwargs.get("date_created"), self._FROM_DICT_MAP
        )
        self.date_modified = time_helper.map_str_to_datetime(
            "date_modified", kwargs.get("date_modified"), self._FROM_DICT_MAP
        )
        self._last_run = time_helper.map_str_to_datetime(
            'last_run', kwargs.get('last_run'), self._FROM_DICT_MAP
        )
        self.owner = (
            User.from_dict(kwargs.get('owner'), self.connection)
            if kwargs.get('owner')
            else None
        )
        self.schedules = (
            [
                Schedule.from_dict(schedule, self._connection)
                for schedule in kwargs.get('schedules')
            ]
            if kwargs.get('schedules')
            else None
        )
        self.contents = (
            [
                Content.from_dict(content, self._connection)
                for content in kwargs.get('contents')
            ]
            if kwargs.get('contents')
            else None
        )
        self.recipients = kwargs.get('recipients')
        self.delivery = (
            Delivery.from_dict(kwargs.get('delivery'))
            if kwargs.get('delivery')
            else None
        )
        self.project_id = project_id
        self._status = kwargs.get('status')
        self._prompts = None

    @method_version_handler('11.3.0000')
    def alter(
        self,  # NOSONAR
        name: str | None = None,
        multiple_contents: bool | None = None,
        allow_delivery_changes: bool | None = None,
        allow_personalization_changes: bool | None = None,
        allow_unsubscribe: bool | None = None,
        send_now: bool = False,
        owner_id: str | None = None,
        schedules: str | list[str] | Schedule | list[Schedule] | None = None,
        contents: Content | list[Content] | None = None,
        recipients: list[str] | list[dict] | None = None,
        delivery: Delivery | dict | None = None,
        delivery_mode: str | None = None,
        custom_msg: str | None = None,
        delivery_expiration_date: str | None = None,
        delivery_expiration_timezone: str | None = None,
        contact_security: bool | None = None,
        filename: str | None = None,
        compress: bool | None = None,
        space_delimiter: str | None = None,
        email_subject: str | None = None,
        email_message: str | None = None,
        email_send_content_as: str | None = None,
        overwrite_older_version: bool | None = None,
        zip_filename: str | None = None,
        zip_password_protect: bool | None = None,
        zip_password: str | None = None,
        file_burst_sub_folder: str | None = None,
        printer_copies: int | None = None,
        printer_range_start: int | None = None,
        printer_range_end: int | None = None,
        printer_collated: bool | None = None,
        printer_orientation: str | None = None,
        printer_use_print_range: bool | None = None,
        cache_cache_type: str | None = None,
        cache_shortcut_cache_format: str | None = None,
        mobile_client_type: str | None = None,
        device_id: str | None = None,
        do_not_create_update_caches: bool | None = None,
        re_run_hl: bool | None = None,
        cache_library_cache_types: list[LibraryCacheTypes | str] = None,
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: str | None = None,
    ):
        """
        Alter subscription.

        Args:
            name (str): name of the subscription,
            multiple_contents (bool, optional): whether multiple contents are
                allowed
            allow_delivery_changes (bool): whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes (bool): whether the recipients can
                personalize the subscription,
            allow_unsubscribe (bool): whether the recipients can unsubscribe
                from the subscription,
            send_now (bool): indicates whether to execute the subscription
                immediately,
            owner_id (str): ID of the subscription owner, by default logged in
                user ID,
            schedules (str | list[str] | Schedule | list[Schedule]):
                Schedules IDs or Schedule objects,
            contents (Content | list[Content]): The content of the subscription.
            recipients (list[str] | list[dict]): list of recipients IDs
                or dicts,
            delivery (Delivery, dict, optional): delivery settings
            delivery_mode (str, enum): the subscription delivery mode [EMAIL,
                FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP, SNAPSHOT,
                PERSONAL_VIEW, SHARED_LINK, UNSUPPORTED],
            custom_msg (str, optional): customized message displayed when
                Subscription has been successfully altered
            delivery_expiration_date (str): expiration date of the subscription,
                format should be yyyy-MM-dd,
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription
            contact_security (bool): whether to use contact security for each
                contact group member,
            filename (str): the filename that will be delivered when
                the subscription is executed,
            compress (bool): whether to compress the file
            space_delimiter (str): space delimiter,
            email_subject (str): email subject associated with the subscription,
            email_message (str): email body of subscription,
            email_send_content_as (str,enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename (str): filename of the compressed content,
            zip_password_protect (bool): whether to password protect zip file,
            zip_password (str): optional password for the compressed file
            file_burst_sub_folder (str): burst sub folder,
            printer_copies (int): the number of copies that should be printed,
            printer_range_start (int): the number indicating the first report
                page that should be printed,
            printer_range_end (int): the number indicating the last report
                page that should be printed,
            printer_collated (bool): whether the printing should be collated,
            printer_orientation (str,enum): [ PORTRAIT, LANDSCAPE ]
            printer_use_print_range (bool): whether print range should be used,
            cache_cache_type (str,enum): [RESERVED, SHORTCUT,
                SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format (str,enum): [RESERVED, JSON, BINARY,
                BOTH]
            mobile_client_type (str,enum): [PHONE, TABLET]
            device_id (str): the mobile target project,
            do_not_create_update_caches (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            re_run_hl (bool): whether subscription will re-run against warehouse
            cache_library_cache_types (list[LibraryCacheTypes | str], optional):
                Set of library cache types, available types can be
                web, android, ios
            cache_reuse_dataset_cache (bool): Whether to reuse dataset cache
            cache_is_all_library_users (bool): Whether for all library users
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """

        # TODO Potentially remove if new subscription types are supported
        cache_library_cache_types = cache_library_cache_types or [LibraryCacheTypes.WEB]
        if self.delivery.mode in [
            'SNAPSHOT',
            'PERSONAL_VIEW',
            'SHARED_LINK',
            'UNSUPPORTED',
        ]:
            helper.exception_handler(
                msg=f'{self.delivery.mode} subscription altering is not supported.',
                exception_type=NotSupportedError,
            )
        # Schedules logic
        schedules = self.__validate_schedules(schedules=schedules)
        if not schedules:
            schedules = [{'id': sch.id} for sch in self.schedules]

        # Content logic
        if contents:
            contents = self.__validate_contents(contents)
        else:
            # Even if contents are not changed, they must be executed with
            # any stored prompt answers
            contents = self.__get_reprompted_contents()

        # Delivery logic
        delivery_expiration_timezone = self.__validate_expiration_time_zone(
            self.connection,
            delivery_expiration_timezone,
        )
        if delivery:
            temp_delivery = (
                Delivery.from_dict(delivery) if isinstance(delivery, dict) else delivery
            )
        else:
            personal_notification_address_id = delivery_personal_notification_address_id

            temp_delivery = self.__change_delivery_properties(
                delivery_mode,
                delivery_expiration_date,
                contact_security,
                email_subject,
                email_message,
                filename,
                compress,
                None,
                zip_filename,
                zip_password,
                zip_password_protect,
                space_delimiter,
                email_send_content_as,
                overwrite_older_version,
                file_burst_sub_folder,
                printer_copies,
                printer_range_start,
                printer_range_end,
                printer_collated,
                printer_orientation,
                printer_use_print_range,
                expiration_time_zone=delivery_expiration_timezone,
                mobile_client_type=mobile_client_type,
                device_id=device_id,
                do_not_create_update_caches=do_not_create_update_caches,
                re_run_hl=re_run_hl,
                cache_type=cache_cache_type,
                shortcut_cache_format=cache_shortcut_cache_format,
                library_cache_types=cache_library_cache_types,
                reuse_dataset_cache=cache_reuse_dataset_cache,
                is_all_library_users=cache_is_all_library_users,
                notification_enabled=delivery_notification_enabled,
                personal_notification_address_id=personal_notification_address_id,
            )
        delivery = temp_delivery.to_dict(camel_case=True)

        # Recipients logic
        recipients = self.__is_val_changed(recipients=recipients)
        recipients = Subscription._validate_recipients(
            self.connection,
            contents,
            recipients,
            self.project_id,
            delivery['mode'],
            self.recipients,
        )

        body = {
            "name": self.__is_val_changed(name=name),
            "allowDeliveryChanges": self.__is_val_changed(
                allow_delivery_changes=allow_delivery_changes
            ),
            "multipleContents": self.__is_val_changed(
                multiple_contents=multiple_contents
            ),
            "allowPersonalizationChanges": self.__is_val_changed(
                allow_personalization_changes=allow_personalization_changes
            ),
            "allowUnsubscribe": self.__is_val_changed(
                allow_unsubscribe=allow_unsubscribe
            ),
            "sendNow": send_now,
            'owner': {
                'id': self.__is_val_changed(nested=self.owner.id, owner_id=owner_id)
            },
            "schedules": schedules,
            "contents": contents,
            "recipients": recipients,
            "delivery": delivery,
        }

        body = helper.delete_none_values(body, recursion=True)

        response = subscriptions.update_subscription(
            self.connection, self.id, self.project_id, body
        )

        if response.ok:
            response = response.json()
            response = helper.camel_to_snake(response)
            self._set_object_attributes(**response)
            if config.verbose:
                msg = f"Updated subscription '{self.name}' with ID: {self.id}."
                msg = custom_msg or msg
                logger.info(msg)

    def __is_val_changed(self, nested=None, **kwargs):
        for key, value in kwargs.items():
            if nested:
                return value if value != nested and value is not None else nested
            current_val = self.__dict__.get(key)
            # if not current_val: we need to get
            return value if value != current_val and value is not None else current_val

    @staticmethod
    def __validate_schedules(
        schedules: str | list[str] | Schedule | list[Schedule] = None,
    ):
        schedules = schedules if isinstance(schedules, list) else [schedules]
        schedules = [
            {'id': schedule.id if isinstance(schedule, Schedule) else schedule}
            for schedule in schedules
            if schedule is not None
        ]
        return schedules

    @staticmethod
    def __validate_contents(
        contents: list[Content | dict] | Content | dict,
    ) -> list[dict]:
        contents = contents if isinstance(contents, list) else [contents]
        content_type_msg = "Contents must be dictionaries or Content objects."
        return [
            (
                content.to_dict(camel_case=True)
                if isinstance(content, Content)
                else (
                    content
                    if isinstance(content, dict)
                    else helper.exception_handler(content_type_msg, TypeError)
                )
            )
            for content in contents
        ]

    def execute(self) -> None:
        """Executes a subscription with given name or GUID for given project."""
        subscriptions.send_subscription(self.connection, self.id, self.project_id)
        if config.verbose:
            logger.info(f"Executed subscription '{self.name}' with ID: {self.id}.")

    @method_version_handler('11.2.0203')
    def delete(self, force: bool = False) -> bool:
        """Delete a subscription. Returns True if deletion was successful.

        Args:
            force (bool, optional): If True, no additional prompt will be shown
                before deleting the subscription. Defaults to False.
        """
        message = (
            f"Are you sure you want to delete subscription "
            f"'{self.name}' with ID: {self.id}? [Y/N]: "
        )
        if not force and input(message) != 'Y':
            return False

        response = subscriptions.remove_subscription(
            self.connection, self.id, self.project_id
        )
        if config.verbose:
            logger.info(f"Deleted subscription '{self.name}' with ID: {self.id}.")
        return response.ok

    @method_version_handler('11.3.0000')
    def available_bursting_attributes(self) -> dict:
        """Get a list of available attributes for bursting feature."""
        contents_bursting = {}
        for content in self.contents:
            response = subscriptions.bursting_attributes(
                self.connection,
                self.project_id,
                content.id,
                content.type.upper(),
            )
            if response.ok:
                contents_bursting[content.id] = response.json()['burstingAttributes']

        return contents_bursting

    @method_version_handler('11.3.0000')
    def available_recipients(self) -> list[dict]:
        """List available recipients for subscription content."""
        body = {"contents": [content.to_dict() for content in self.contents]}
        delivery_type = self.delivery.mode
        response = subscriptions.available_recipients(
            self.connection, self.project_id, body, delivery_type
        )

        if response.ok and config.verbose:
            return response.json()['recipients']

    @method_version_handler('11.3.0000')
    def add_recipient(
        self,
        recipients: list[dict] | dict | list[str] | str = None,
        recipient_id: str | None = None,
        recipient_type: str | None = None,
        recipient_include_type: str = 'TO',
    ):
        """Adds recipient to subscription. You can either specify id, type and
        include_type of single recipient, or just pass recipients list as a
        list of dictionaries.

        Note:
            When providing recipient ID remember to also provide its type.

        Args:
            recipients: list of ids or dicts containing recipients, dict format:
                {"id": recipient_id,
                 "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                         "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                 "includeType": "TO" / "CC" / "BCC"}
            recipient_id: id of the recipient
            recipient_type: type of the recipient
            recipient_include_type: include type of the recipient one of the
                following: "TO" / "CC" / "BCC", by default "TO"
        """
        if recipients:
            recipients = recipients if isinstance(recipients, list) else [recipients]
        else:
            recipients = []
        if recipient_id and recipient_type:
            recipients.append(
                {
                    "id": recipient_id,
                    "type": recipient_type,
                    "includeType": recipient_include_type,
                }
            )
        elif (recipients == [] and recipient_id is None) or (
            len(recipients) >= 1 and recipient_id
        ):
            msg = (
                "Specify either a recipient ID, type and include type or pass "
                "recipients dictionaries"
            )
            helper.exception_handler(msg, ValueError)
        elif recipients == [] and recipient_id and recipient_type is None:
            msg = "When providing recipient ID remember to also provide its type."
            helper.exception_handler(msg, ValueError)

        all_recipients = self.recipients.copy()
        ready_recipients = self.__prepare_recipients(recipients)

        ready_recipients = self._validate_recipients(
            connection=self.connection,
            contents=self.contents,
            recipients=ready_recipients,
            project_id=self.project_id,
            delivery_mode=self.delivery.mode,
            current_recipients=self.recipients,
        )

        if ready_recipients:
            all_recipients.extend(ready_recipients)
            self.alter(recipients=all_recipients)
        elif config.verbose:
            logger.info('No recipients were added to the subscription.')

    @method_version_handler('11.3.0000')
    def remove_recipient(self, recipients: list[str] | list[dict]):
        """Removes recipient from given subscription in given project.

        Args:
            recipients: list of ids or dicts containing recipients, dict format:
                {"id": recipient_id,
                 "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                         "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                 "includeType": "TO" / "CC" / "BCC"}
        """
        all_recipients = self.recipients
        recipients = recipients if isinstance(recipients, list) else [recipients]
        existing_recipients = [rec['id'] for rec in self.recipients]

        if len(self.recipients) == 1:
            helper.exception_handler(
                "Subscription must have at last one recipient. Add new recipient "
                "before removing.",
                exception_type=ValueError,
            )
        for recipient in recipients:
            rec_id = recipient['id'] if isinstance(recipient, dict) else recipient
            if rec_id not in existing_recipients:
                helper.exception_handler(
                    f"{rec_id} is not a recipient of subscription", UserWarning
                )
            else:
                all_recipients = [rec for rec in all_recipients if rec['id'] != rec_id]
        if len(all_recipients) == 0:
            helper.exception_handler(
                "You cannot remove all existing recipients. Add new recipient before "
                "removing.",
                exception_type=ValueError,
            )
        elif len(self.recipients) - len(all_recipients) > 0:
            self.alter(recipients=all_recipients)
        elif len(self.recipients) == len(all_recipients) and config.verbose:
            logger.info('No recipients were removed from the subscription.')

    def __prepare_recipients(self, recipients):
        existing_recipients = [rec['id'] for rec in self.recipients]
        ready_recipients = []

        def __already_recipient(recipient):
            if recipient in existing_recipients:
                helper.exception_handler(
                    f"{recipient} is already a recipient of subscription", UserWarning
                )
            else:
                ready_recipients.append(recipient)

        for recipient in recipients:
            not_dict_msg = """Wrong recipient format, expected format is
                              {"id": recipient_id,
                               "type": "CONTACT_GROUP" / "USER_GROUP" / "CONTACT" /
                                       "USER" / "PERSONAL_ADDRESS" / "UNSUPPORTED"
                               "includeType": "TO" / "CC" / "BCC" (optional)
                              }"""
            if isinstance(recipient, dict):
                if list(recipient.keys()) in [
                    ['id', 'type', 'includeType'],
                    ['id', 'type'],
                ]:
                    __already_recipient(recipient['id'])
                else:
                    helper.exception_handler(not_dict_msg, TypeError)
            if isinstance(recipient, str):
                __already_recipient(recipient)
        return ready_recipients

    def __change_delivery_properties(  # NOSONAR
        self,  # NOSONAR
        mode: str | None = None,
        expiration: str | None = None,
        contact_security: bool | None = None,
        subject: str | None = None,
        message: str | None = None,
        filename: str | None = None,
        compress: bool | None = None,
        zip_settings: ZipSettings | None = None,
        zip_filename: str | None = None,
        zip_password: str | None = None,
        zip_password_protect: bool | None = None,
        space_delimiter: str | None = None,
        send_content_as: SendContentAs | None = None,
        overwrite_older_version: bool | None = None,
        burst_sub_folder: str | None = None,
        copies: int | None = None,
        range_start: int | None = None,
        range_end: int | None = None,
        collated: bool | None = None,
        orientation: Orientation | None = None,
        use_print_range: bool | None = None,
        expiration_time_zone: str | None = None,
        cache_type: CacheType | None = None,
        shortcut_cache_format: ShortcutCacheFormat | None = None,
        mobile_client_type: ClientType | None = None,
        device_id: str | None = None,
        do_not_create_update_caches: bool | None = None,
        re_run_hl: bool | None = None,
        library_cache_types: list[LibraryCacheTypes | str] = None,
        reuse_dataset_cache: bool = False,
        is_all_library_users: bool = False,
        notification_enabled: bool = False,
        personal_notification_address_id: str | None = None,
    ):
        library_cache_types = library_cache_types or [LibraryCacheTypes.WEB]
        func = self.__change_delivery_properties
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        # create dict of properties to be changed
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]

        # 'zip_settings' is 'zip' in object
        if 'zip_settings' in properties:
            properties['zip'] = properties.pop('zip_settings')

        obj_dict = self.delivery.VALIDATION_DICT
        obj_mode_dict = self.delivery.__dict__[
            self.delivery.mode.lower()
        ].VALIDATION_DICT
        obj_mode_zip_dict = {
            "zip_filename": str,
            "zip_password": str,
            "zip_password_protect": bool,
        }

        if properties:  # at this point only not None values in properties
            for key, value in properties.items():
                if (
                    key in obj_dict
                ):  # Highest level parameters. Mainly mode type and object
                    # Change mode or set attr if it's not mode param
                    if not self.delivery.change_mode(key, value):
                        self.delivery.__setattr__(key, value)
                elif (
                    key in obj_mode_dict
                ):  # parameters of a particular mode e.g. of email
                    helper.rsetattr(
                        self.delivery, f'{self.delivery.mode.lower()}.{key}', value
                    )
                elif key in obj_mode_zip_dict:  # zip settings
                    key = key[4:]
                    if not helper.rgetattr(
                        self.delivery, f'{self.delivery.mode.lower()}.zip', None
                    ):
                        helper.rsetattr(
                            self.delivery,
                            f'{self.delivery.mode.lower()}.zip',
                            ZipSettings(),
                        )
                    helper.rsetattr(
                        self.delivery, f'{self.delivery.mode.lower()}.zip.{key}', value
                    )
        return self.delivery

    @classmethod
    def from_dict(
        cls,
        source: dict[str, Any],
        connection: "Connection" = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ) -> "Subscription":
        """Initialize Subscription object from dictionary.

        Args:
            source: (dict) A dictionary containing subscription data.
            connection: (Connection) A Strategy connection object.
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name

        Returns:
            A Subscription object.
        """

        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id or source.get('project_id'),
            project_name,
        )

        _source = {
            **source,
            "project_id": proj_id,
        }
        obj: Subscription = super().from_dict(_source, connection)

        return obj

    @staticmethod
    def _check_is_content_prompted(
        connection: 'Connection', content: dict, project_id: str | None = None
    ) -> bool:
        def check_prompts(
            inst_id: str,
            get_prompts_func: Callable,
            get_status_func: Callable,
            c_id: str,
            c_type: str,
        ) -> bool:
            prompts = get_response_json(get_prompts_func())
            if any(prompt.get('required') for prompt in prompts):
                if not inst_id:
                    err_msg = (
                        f"The content of type '{c_type.capitalize()}' with ID '{c_id}' "
                        f"has unanswered prompts. "
                        f"Please use the `{c_type.capitalize()}.answer_prompts()` "
                        f"method to answer them. "
                        f"After that, pass the `{c_type.capitalize()}.instance_id` "
                        f"to the subscription content."
                    )
                    raise ValueError(err_msg)

                status = get_response_json(get_status_func()).get('status')
                if status == 2:
                    err_msg = (
                        f"{c_type.capitalize()}.instance_id: {inst_id} "
                        f"has unanswered prompts. "
                        f"You can use the `{c_type.capitalize()}.answer_prompts()`"
                        f" method to answer them."
                    )
                    raise ValueError(err_msg)

                return True
            return False

        content_type = content.get('type')
        content_id = content.get('id')
        instance_id = (
            content.get('personalization', {}).get('prompt', {}).get('instanceId')
        )

        common_args = {
            'connection': connection,
        }

        if content_type == 'report':
            common_args.update({'report_id': content_id})
            return check_prompts(
                inst_id=instance_id,
                get_prompts_func=partial(
                    reports.get_report_prompts,
                    **common_args,
                    closed=False,
                ),
                get_status_func=partial(
                    reports.get_report_status,
                    **common_args,
                    project_id=project_id,
                    instance_id=instance_id,
                ),
                c_id=content_id,
                c_type=content_type,
            )
        elif content_type in ['document', 'dossier']:
            common_args.update(
                {
                    'document_id': content_id,
                    'project_id': project_id,
                }
            )
            return check_prompts(
                inst_id=instance_id,
                get_prompts_func=partial(
                    documents.get_prompts,
                    **common_args,
                    closed=False,
                ),
                get_status_func=partial(
                    documents.get_document_status,
                    **common_args,
                    instance_id=instance_id,
                ),
                c_id=content_id,
                c_type=content_type,
            )
        return False

    @classmethod
    @method_version_handler('11.3.0000')
    def __create(  # NOSONAR
        cls,  # NOSONAR
        connection: Connection,
        name: str,
        contents: Content | dict,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        multiple_contents: bool | None = None,
        allow_delivery_changes: bool | None = None,
        allow_personalization_changes: bool | None = None,
        allow_unsubscribe: bool | None = None,
        send_now: bool | None = None,
        owner_id: str | None = None,
        schedules: str | list[str] | Schedule | list[Schedule] | None = None,
        recipients: list[dict] | list[str] | None = None,
        delivery: Delivery | dict | None = None,
        delivery_mode: str = Delivery.DeliveryMode.EMAIL,
        delivery_expiration_date: str | None = None,
        delivery_expiration_timezone: str | None = None,
        contact_security: bool = True,
        filename: str | None = None,
        compress: bool = False,
        space_delimiter: str | None = None,
        email_subject: str | None = None,
        email_message: str | None = None,
        email_send_content_as: str = SendContentAs.DATA,
        overwrite_older_version: bool = False,
        zip_filename: str | None = None,
        zip_password_protect: bool | None = None,
        zip_password: str | None = None,
        file_burst_sub_folder: str | None = None,
        printer_copies: int = 1,
        printer_range_start: int = 0,
        printer_range_end: int = 0,
        printer_collated: bool = True,
        printer_orientation: str = Orientation.PORTRAIT,
        printer_use_print_range: bool = False,
        cache_cache_type: CacheType | str = CacheType.RESERVED,
        cache_shortcut_cache_format: (
            ShortcutCacheFormat | str
        ) = ShortcutCacheFormat.RESERVED,
        mobile_client_type: str = ClientType.PHONE,
        device_id: str | None = None,
        do_not_create_update_caches: bool = True,
        re_run_hl: bool = True,
        cache_library_cache_types: list[LibraryCacheTypes | str] = None,
        cache_reuse_dataset_cache: bool = False,
        cache_is_all_library_users: bool = False,
        delivery_notification_enabled: bool = False,
        delivery_personal_notification_address_id: str | None = None,
    ):
        """Creates a subscription Create_Subscription_Outline.

        Args:
            connection (Connection): a Strategy One connection object
            name (str): name of the subscription,
            contents (Content): The content settings.
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            multiple_contents (bool, optional): whether multiple contents are
                allowed
            allow_delivery_changes (bool): whether the recipients can change
                the delivery of the subscription,
            allow_personalization_changes (bool): whether the recipients can
                personalize the subscription,
            allow_unsubscribe (bool): whether the recipients can unsubscribe
                from the subscription,
            send_now (bool): indicates whether to execute the subscription
                immediately,
            owner_id (str): ID of the subscription owner, by default logged in
                user ID,
            schedules (str | list[str] | Schedule | List[Schedule]):
                Schedules IDs or Schedule objects,
            recipients (list[dict], list[str]): list of recipients IDs or dicts,
            delivery (Delivery | dict): delivery object or dict
            delivery_mode (str, enum): the subscription delivery mode [EMAIL,
                FILE, PRINTER, HISTORY_LIST, CACHE, MOBILE, FTP, SNAPSHOT,
                PERSONAL_VIEW, SHARED_LINK, UNSUPPORTED],
            delivery_expiration_date (str, optional): expiration date of the
                subscription, format should be yyyy-MM-dd,
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription, example value 'Europe/London'
            contact_security (bool): whether to use contact security for each
                contact group member,
            filename (str): the filename that will be delivered when
                the subscription is executed,
            compress (bool): whether to compress the file,
            space_delimiter (str): space delimiter,
            email_subject (str): email subject associated with the subscription,
            email_message (str): email body of subscription,
            email_send_content_as (str,enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename (str): filename of the compressed content,
            zip_password_protect (bool): whether to password protect zip file,
            zip_password (str): optional password for the compressed file,
            file_burst_sub_folder (str): burst sub folder,
            printer_copies (int): the number of copies that should be printed,
            printer_range_start (int): the number indicating the first report
                page that should be printed,
            printer_range_end (int): the number indicating the last report
                page that should be printed,
            printer_collated (bool): whether the printing should be collated,
            printer_orientation (str,enum): [ PORTRAIT, LANDSCAPE ]
            printer_use_print_range (bool): whether print range should be used,
            cache_cache_type (str,enum): [RESERVED, SHORTCUT,
                SHORTCUTWITHBOOKMARK]
            cache_shortcut_cache_format (str,enum):
                [RESERVED, JSON, BINARY, BOTH]
            mobile_client_type (str,enum): [PHONE, TABLET]
            device_id (str): the mobile target project,
            do_not_create_update_caches (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            re_run_hl (bool): whether subscription will re-run against warehouse
            cache_library_cache_types (list[LibraryCacheTypes | str], optional):
                Set of library cache types, available types can be
                web, android, ios
            cache_reuse_dataset_cache (bool): Whether to reuse dataset cache
            cache_is_all_library_users (bool): Whether for all library users
            delivery_notification_enabled (bool): Whether notification is
                enabled, notification applies to cache
            delivery_personal_notification_address_id (str, optional):
                Notification details
        """
        cache_library_cache_types = cache_library_cache_types or [LibraryCacheTypes.WEB]

        cache_cache_type = (
            LegacyCacheType[cache_cache_type.name]
            if not is_server_min_version(connection, '11.3.0100')
            else CacheType[cache_cache_type.name]
        )

        name = (
            name
            if len(name) <= 255
            else helper.exception_handler(
                "Name too long. Max name length is 255 characters."
            )
        )
        proj_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
        )

        if not schedules:
            msg = "Please specify 'schedules' parameter."
            helper.exception_handler(msg)

        schedules = cls.__validate_schedules(schedules=schedules)

        # Content logic
        contents = cls.__validate_contents(contents)

        for content in contents:
            cls._check_is_content_prompted(connection, content, proj_id)

        # Delivery logic
        delivery_expiration_timezone = cls.__validate_expiration_time_zone(
            connection,
            delivery_expiration_timezone,
        )
        if delivery:
            temp_delivery = (
                Delivery.from_dict(delivery) if isinstance(delivery, dict) else delivery
            )
        else:
            personal_notification_address_id = delivery_personal_notification_address_id

            temp_delivery = Delivery(
                delivery_mode,
                delivery_expiration_date,
                contact_security,
                email_subject,
                email_message,
                filename,
                compress,
                None,
                zip_password,
                zip_password_protect,
                space_delimiter,
                email_send_content_as,
                overwrite_older_version,
                file_burst_sub_folder,
                printer_copies,
                printer_range_start,
                printer_range_end,
                printer_collated,
                printer_orientation,
                printer_use_print_range,
                mobile_client_type,
                device_id,
                do_not_create_update_caches,
                re_run_hl,
                cache_type=cache_cache_type,
                shortcut_cache_format=cache_shortcut_cache_format,
                library_cache_types=cache_library_cache_types,
                reuse_dataset_cache=cache_reuse_dataset_cache,
                is_all_library_users=cache_is_all_library_users,
                notification_enabled=delivery_notification_enabled,
                personal_notification_address_id=personal_notification_address_id,
                expiration_time_zone=delivery_expiration_timezone,
            )
        delivery = temp_delivery.to_dict(camel_case=True)

        # Recipients logic
        recipients = cls._validate_recipients(
            connection, contents, recipients, proj_id, delivery['mode']
        )

        # Create body
        body = {
            "multipleContents": len(contents) > 1,
            "name": name,
            "allowDeliveryChanges": allow_delivery_changes,
            "allowPersonalizationChanges": allow_personalization_changes,
            "allowUnsubscribe": allow_unsubscribe,
            "sendNow": send_now,
            "owner": {"id": owner_id},
            "schedules": schedules,
            "contents": contents,
            "recipients": recipients,
            "delivery": delivery,
        }

        body = helper.delete_none_values(body, recursion=True)
        response = subscriptions.create_subscription(connection, proj_id, body)
        unpacked_response = response.json()
        if config.verbose:
            logger.info(
                f"Created subscription '{name}' with ID: '{unpacked_response['id']}'."
            )
        return cls.from_dict(unpacked_response, connection, proj_id)

    @staticmethod
    def _validate_recipients(
        connection: "Connection",
        contents: list[Content | dict],
        recipients: list[str] | list[dict] | str,
        project_id: str,
        delivery_mode: str,
        current_recipients: list[dict] | None = None,
    ):
        def __not_available(recipient):
            if recipient not in available_recipients_ids:
                msg = (
                    f"'{recipient}' is not a valid recipient ID for selected content "
                    "and delivery mode. Available recipients: \n"
                    f"{pformat(available_recipients, indent=2)}"
                )
                helper.exception_handler(msg, ValueError)

            rec = helper.filter_list_of_dicts(available_recipients, id=recipient)
            formatted_recipients.append(rec[0])

        recipients = recipients or []
        recipients = recipients if isinstance(recipients, list) else [recipients]
        body = {
            "contents": [
                cont.to_dict() if isinstance(cont, Content) else cont
                for cont in contents
            ]
        }
        available_recipients = subscriptions.available_recipients(
            connection, project_id, body, delivery_mode
        ).json()['recipients'] + (current_recipients or [])
        available_recipients_ids = [rec['id'] for rec in available_recipients]
        # Format recipients list if needed
        formatted_recipients = []
        if recipients:
            for recipient in recipients:
                if isinstance(recipient, dict):
                    __not_available(recipient['id'])
                elif isinstance(recipient, str):
                    __not_available(recipient)
                else:
                    recipient_type = type(recipient)

                    helper.exception_handler(
                        "Recipients must be a dictionaries or a strings, "
                        f"not a {recipient_type}",
                        exception_type=TypeError,
                    )
        return formatted_recipients

    @staticmethod
    def __validate_expiration_time_zone(connection, expiration_time_zone):
        if connection._iserver_version < '11.3.1000' and expiration_time_zone:
            if config.verbose:
                logger.info(
                    'delivery_expiration_timezone argument is available from '
                    'iServer Version 11.3.1000'
                )
        else:
            return expiration_time_zone

    def _is_valid_delivery_mode(self) -> bool:
        valid_modes = [
            'EMAIL',
            'HISTORY_LIST',
            'CACHE',
            'FTP',
            'FILE',
            'MOBILE',
        ]
        # Get subscription delivery mode from API to avoid max depth recursion
        sub_mode = (
            subscriptions.get_subscription(
                connection=self.connection,
                subscription_id=self.id,
                project_id=self.project_id,
            )
            .json()
            .get('delivery', {})
            .get('mode')
        )
        return sub_mode in valid_modes

    def __get_reprompted_contents(self) -> list[dict]:
        """Re-prompt contents to obtain their representation for alter().
        Returns:
            list[dict]: list of content dicts to be included in REST payload
        """

        from mstrio.modeling.prompt import Prompt
        from mstrio.project_objects.document import Document
        from mstrio.project_objects.report import Report

        contents: list[dict] = [cont.to_dict() for cont in self.contents]
        if not any(ct.personalization.prompt for ct in self.contents):
            return contents
        prompts_data = subscriptions.get_subscription_prompts(
            self.connection,
            self.id,
            self.project_id,
        ).json()['prompts']
        for content in contents:
            content_type = content.get('type')
            dict_update = {
                'enabled': True,
            }
            if content_type == 'report':
                rep = Report(self.connection, content['id'])
                rep.answer_prompts(Prompt.bulk_from_dict(prompts_data), True)
                dict_update['instanceId'] = rep.instance_id
            elif content_type in ['document', 'dossier']:
                doc = Document(self.connection, content['id'])
                doc.answer_prompts(prompts_data, True)
                dict_update['instanceId'] = doc.instance_id
            content['personalization'].update({'prompt': dict_update})
        return contents

    @property
    @method_version_handler(version='11.4.0600')
    def status(self) -> SubscriptionStatus | None:
        if not self._is_valid_delivery_mode():
            msg = (
                f"Subscription with delivery mode: "
                f"{self.delivery.mode} is not supported."
            )
            raise NotSupportedError(msg)
        self.fetch('status')
        return self._status

    @property
    @method_version_handler(version='11.4.0600')
    def last_run(self) -> datetime | None:
        if not self._is_valid_delivery_mode():
            msg = (
                f"Subscription with delivery mode: "
                f"{self.delivery.mode} is not supported."
            )
            raise NotSupportedError(msg)
        return self._last_run

    def fetch(self, attr: str | None = None) -> None:
        if not self._is_valid_delivery_mode():
            self._API_GETTERS = {
                k: v
                for k, v in self._API_GETTERS.items()
                if k not in ['status', 'last_run']
            }
        super().fetch(attr)

    def answer_prompts(
        self,
        prompt_answers: list["Prompt"],
        force: bool = False,
    ) -> bool:
        """Answer prompts of the object.

        Args:
            prompt_answers (list[Prompt]): List of Prompt class objects
                answering the prompts of the object.
            force (bool): If True, then the object's existing prompt will be
                overwritten by ones from the prompt_answers list, and additional
                input from the user won't be asked. Otherwise, the user will be
                asked for input if the prompt is not answered, or if prompt was
                already answered.

        Returns:
            bool: True if prompts were answered successfully, False otherwise.
        """
        from mstrio.project_objects import Document, Report

        new_contents = []
        for content_obj in self.contents:
            selected_obj = None
            if content_obj.type == Content.Type.REPORT:
                selected_obj = Report(self.connection, content_obj.id)
            elif content_obj.type in [
                Content.Type.DOCUMENT,
                Content.Type.DASHBOARD,
            ]:
                selected_obj = Document(self.connection, content_obj.id)
            else:
                raise NotSupportedError(
                    f"Answering prompts is not supported for content type "
                    f"'{content_obj.type}'."
                )

            selected_obj.answer_prompts(prompt_answers, force=force)
            new_contents.append(
                Content(
                    id=selected_obj.id,
                    type=content_obj.type,
                    personalization=Content.Properties(
                        format_type=content_obj.personalization.format_type,
                        prompt=Content.Properties.Prompt(
                            enabled=True,
                            instance_id=selected_obj.instance_id,
                        ),
                    ),
                )
            )

        self.alter(contents=new_contents)
        return True

    @property
    @method_version_handler('11.5.0900')
    def prompts(self) -> dict:
        """Prompts of the report."""
        if self._prompts is None:
            prompts = (
                subscriptions.get_subscription_prompts(
                    connection=self.connection,
                    subscription_id=self.id,
                    project_id=self.project_id,
                )
                .json()
                .get('prompts', [])
            )
            self._prompts = [
                Prompt.from_dict(source=prompt, connection=self.connection)
                for prompt in prompts
            ]
        return self._prompts



import inspect
import logging
from collections import defaultdict
from collections.abc import Callable
from pprint import pformat

logger = logging.getLogger(__name__)


def _get_function_signature_as_formatted_str(func: Callable) -> str:
    return pformat(
        tuple(
            str(v)
            for v in inspect.signature(func).parameters.values()
            if v.name != 'self'
        )
    )


def what_can_i_do_with(source) -> None:
    logger.info('\n{mstrio-py WHAT CAN I DO WITH helper info}\n')

    module = inspect.getmodule(source)
    is_callable = callable(source)
    is_builtin = inspect.isbuiltin(source) or not module or inspect.isbuiltin(module)
    valid = bool(source) and not is_builtin

    if not valid:
        raise ValueError(
            'The provided source is neither callable nor an instance of '
            'a class. Please provide a reference to a callable object, '
            'like a class, a function or a method or to an instance of a class.',
        )

    is_mstrio_based: bool = not is_builtin and (
        module.__name__.startswith('mstrio') if module else False
    )

    if not is_callable:  # aka some class instance
        cls_name = source.__class__.__name__
        is_module = cls_name == 'module'
        if is_module:
            cls_name = source.__name__

        if not is_mstrio_based:
            logger.info(
                f'Artefact "{cls_name}" does not originate from '
                'mstrio-py library. Information below may not be fully accurate. '
                'Please refer to the official documentation of the source library.\n',
            )

        if is_module:
            logger.info(
                f'You have imported a module "{cls_name}". You can access '
                f'its contents using `{cls_name.split(".")[-1]}.<attr>`. '
                'Learn more about the module in documentation.\n',
            )
        else:
            logger.info(
                f'You have initialized an instance of the "{cls_name}" class. '
                'Preferably, you stored a reference to it in a variable, for example '
                f'like this: `<some_var> = {cls_name}(<args>)`',
            )

            # group all available props and methods
            kinds = defaultdict(list)
            for opt in inspect.classify_class_attrs(source.__class__):
                if not opt.name.startswith('_'):
                    kinds[opt.kind].append(opt.name)

            if 'property' in kinds:
                logger.info(
                    '\nPROPERTIES:\n'
                    'You can access a value stored under any of those via: '
                    '`<some_var>.<property>`',
                )
                logger.info(kinds['property'])
                del kinds['property']

            if 'method' in kinds:
                logger.info(
                    '\nMETHODS:\n'
                    'You can call any of the methods using '
                    '`<some_var>.<method>(<args>)` or you can learn more using '
                    '`what_can_i_do_with(<some_var>.<method>)` '
                    '(<method> without brackets!).',
                )
                logger.info(kinds['method'])
                del kinds['method']

            if 'class method' in kinds:
                logger.info(
                    '\nCLASS METHODS:\n'
                    'Those are methods that behave similarly to regular methods, '
                    'however they are not necessarily bound to any instance. '
                    'You can call them like any method above, or using: \n'
                    f'  {cls_name}.<class_method>(<instance>, *<args>)',
                )
                logger.info(kinds['class method'])
                del kinds['class method']

            if 'static method' in kinds:
                logger.info(
                    '\nSTATIC METHODS:\n'
                    'Those have nothing to do with <some_var> instance. '
                    'You can call them directly via '
                    f'`{cls_name}.<static_method>(<args>)` or learn more using '
                    f'`what_can_i_do_with({cls_name}.<static_method>)` '
                    '(<static_method> without brackets!).',
                )
                logger.info(kinds['static method'])
                del kinds['static method']

            if kinds:
                logger.info(
                    '\nOTHERS:\n'
                    'There are some other non-standard attributes. '
                    'Learn more about them in documentation.',
                )
                for kind, names in kinds.items():
                    logger.info(f'\n  {kind.upper()}:\n  {names}')
    else:  # a callable, like a function or a class instance method
        artefact_name = source.__name__
        is_class_factory = inspect.isclass(source) and hasattr(source, '__init__')

        # check if `source` is a method of a class
        is_class_method = hasattr(source, '__self__')

        if not is_mstrio_based:
            logger.info(
                f'Artefact "{artefact_name}" does not originate from mstrio-py '
                'library. Information below may not be fully accurate. '
                'Please refer to the official documentation of the source library.\n',
            )

        if is_class_factory:
            src = source.__init__
            logger.info(
                f'"{artefact_name}" is a module class. '
                'You can initialize an instance of it using '
                f'`<some_var> = {artefact_name}(<args>)`.\n',
            )
        else:
            src = source
            cls_name = source.__self__.__class__.__name__ if is_class_method else None
            is_static_method = False
            if cls_name == 'type':
                is_static_method = True
                # static method: we need to reassign name to a level higher
                cls_name = source.__self__.__name__

            if not is_class_method:
                msg = (
                    f'"{artefact_name}" is a function. You can call it using '
                    f'`{artefact_name}(<args>)` (assign to a variable if the call '
                    'result is needed).\n'
                )
            else:
                if is_static_method:
                    msg = (
                        f'"{artefact_name}" is a static method on {cls_name} class. '
                        f'You can call it using `{cls_name}.{artefact_name}(<args>)` '
                        '(assign to a variable if the call result is needed).\n'
                    )
                else:
                    msg = (
                        f'"{artefact_name}" is a method on {cls_name} class instance. '
                        f'You can call it using `<some_var>.{artefact_name}(<args>)` '
                        f'where <some_var> is a variable you store {cls_name} '
                        'instance in (assign to a variable if the call '
                        'result is needed).\n'
                    )

            logger.info(msg)

        logger.info('DOCSTRING:\n' + source.__doc__) if source.__doc__ else None
        logger.info('\nCALL SIGNATURE:')
        logger.info(_get_function_signature_as_formatted_str(src))

    logger.info('\n{mstrio-py WHAT CAN I DO WITH helper ended}\n')



from enum import auto

from mstrio.utils.enum_helper import AutoName


class DaysOfWeek(AutoName):
    MONDAY = auto()
    TUESDAY = auto()
    WEDNESDAY = auto()
    THURSDAY = auto()
    FRIDAY = auto()
    SATURDAY = auto()
    SUNDAY = auto()
    NONE = None



import logging
from dataclasses import dataclass, field
from enum import auto
from typing import Any, Optional, TypeVar

from mstrio.connection import Connection
from mstrio.helpers import NotSupportedError
from mstrio.modeling import ObjectSubType, SchemaObjectReference
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, camel_to_snake, snake_to_camel
from mstrio.utils.vldb_mixin import ResolvedLocation

logger = logging.getLogger(__name__)

T = TypeVar("T")


@dataclass
class AttributeJoinType(Dictable):
    """Class representation of single joining operation for attribute.

    Attributes:
        attribute (SchemaObjectReference): Reference to the attribute used
            for joining.
        no_participation_in_preserve_row (bool): Value corresponding to the
            targeted VLDB setting of given attribute.
        resolved_location (ResolvedLocation, optional): Type of location where
            value of related VLDB setting was last time resolved.
        is_inherited (bool, optional): Whether the related VLDB setting value
            is inherited from other objects or not. If False value is set on
            the object. If True, value is inherited from other objects.
        next_no_participation_in_preserve_row (bool, optional): Previously set
            value of attribute targeted VLDB setting.
        next_resolved_location (ResolvedLocation, optional): Previously set
            resolved_location.
    """

    _FROM_DICT_MAP = {
        'attribute': SchemaObjectReference.from_dict,
        'resolved_location': ResolvedLocation,
        'next_resolved_location': ResolvedLocation,
    }

    attribute: SchemaObjectReference
    no_participation_in_preserve_row: bool
    resolved_location: ResolvedLocation = ResolvedLocation.DEFAULT
    is_inherited: bool = True
    next_no_participation_in_preserve_row: bool | None = None
    next_resolved_location: ResolvedLocation | None = None


class JoinType(AutoName):
    """Enumeration constant indicating type of metric joining operation."""

    INNER = auto()
    OUTER = auto()


@dataclass
class MetricJoinType(Dictable):
    """Class representation of single joining operation for metric.

    Attributes:
        metric (SchemaObjectReference): Reference to the metric used for
            joining.
        join_type (JoinType): Value corresponding to the targeted VLDB setting
            of given metric.
        resolved_location (ResolvedLocation, optional): Type of location where
            value of related VLDB setting was last time resolved.
        is_inherited (bool, optional): Whether the related VLDB setting value
            is inherited from other objects or not. If False value is set on
            the object. If True, value is inherited from other objects.
        next_join_type (JoinType, optional): Previously set value of metric
            targeted VLDB setting.
        next_resolved_location (ResolvedLocation, optional): Previously set
            resolved_location.
    """

    _FROM_DICT_MAP = {
        'metric': SchemaObjectReference.from_dict,
        'join_type': JoinType,
        'resolved_location': ResolvedLocation,
        'next_join_type': JoinType,
        'next_resolved_location': ResolvedLocation,
    }

    metric: SchemaObjectReference
    join_type: JoinType
    resolved_location: ResolvedLocation = ResolvedLocation.DEFAULT
    is_inherited: bool = True
    next_join_type: JoinType | None = None
    next_resolved_location: ResolvedLocation | None = None


@dataclass
class AdvancedProperties(Dictable):
    """Class representation of advanced properties containing joining
    information.

    Attributes:
        metric_join_types (list[MetricJoinType], optional): List containing
            joining information for each metric that appears in the template.
        attribute_join_types (list[AttributeJoinType], optional): List
            containing joining information for each attribute that appears in
            the template.
    """

    _KEEP_CAMEL_CASE = [
        'metricJoinTypes',
        'attributeJoinTypes',
        'metric_join_types',
        'attribute_join_types',
    ]

    @staticmethod
    def _parse_metric(source, connection):
        return [
            MetricJoinType.from_dict(value, connection) for value in source.values()
        ]

    @staticmethod
    def _parse_attribute(source, connection):
        return [
            AttributeJoinType.from_dict(value, connection) for value in source.values()
        ]

    _FROM_DICT_MAP = {
        'metric_join_types': _parse_metric,
        'attribute_join_types': _parse_attribute,
    }

    metric_join_types: list[MetricJoinType] = field(default_factory=list)
    attribute_join_types: list[AttributeJoinType] = field(default_factory=list)

    def to_dict(self, camel_case: bool = True) -> dict:
        result = {
            'metric_join_types': (
                {
                    value.metric.object_id: value.to_dict(camel_case)
                    for value in self.metric_join_types
                }
                if self.metric_join_types
                else None
            ),
            'attribute_join_types': (
                {
                    value.attribute.object_id: value.to_dict(camel_case)
                    for value in self.attribute_join_types
                }
                if self.attribute_join_types
                else None
            ),
        }
        return (
            snake_to_camel(result, whitelist=self._KEEP_CAMEL_CASE)
            if camel_case
            else result
        )


class TemplateUnitType(AutoName):
    """Enumeration constant indicating type of single template unit."""

    ATTRIBUTE = auto()
    DIMENSION = auto()
    METRICS = auto()
    CUSTOM_GROUP = auto()
    CONSOLIDATION = auto()
    PROMPT = auto()
    RAW_UNIT = auto()


@dataclass
class TemplateUnit(Dictable):
    """Class representation of single template unit."""

    _TYPE = None

    @staticmethod
    def dispatch(source, connection: Connection | None = None) -> type['TemplateUnit']:
        data = source.copy()
        unit_type = TemplateUnitType(data.pop('type'))

        if unit_type == TemplateUnitType.ATTRIBUTE:
            cls = AttributeTemplateUnit
        elif unit_type == TemplateUnitType.METRICS:
            cls = MetricTemplateUnit
        else:
            raise NotSupportedError(
                f"Unit type: '{unit_type}' is not supported for Template class."
            )

        return cls.from_dict(data, connection)

    def to_dict(self, camel_case: bool = True) -> dict:
        return {
            **super().to_dict(camel_case),
            'type': self._TYPE.value,
        }


@dataclass
class AttributeFormReference(Dictable):
    """Class representation of single attribute form.

    Attributes:
        id (str): ID of the referenced attribute.
        name (str, optional): Name of attribute.
    """

    id: str
    name: str | None = None


@dataclass
class AttributeTemplateUnit(TemplateUnit):
    """Class representation of single attribute template unit.

    Attributes:
        id (str): ID of the referenced attribute.
        name (str, optional): Name of attribute.
        alias (str, optional): Alias for attribute.
        forms (list[AttributeFormReference], optional): Referenced
            attribute forms.
    """

    _TYPE = TemplateUnitType.ATTRIBUTE
    _FROM_DICT_MAP = {
        'forms': [AttributeFormReference.from_dict],
    }

    id: str
    name: str | None = None
    alias: str | None = None
    forms: list[AttributeFormReference] | None = None


@dataclass
class MetricElement(Dictable):
    """Class representation of single metric element.

    Attributes:
        id (str): ID of the referenced metric.
        name (str, optional): Name of metric.
        alias (str, optional): Alias for metric.
    """

    _TYPE = ObjectSubType.METRIC

    id: str
    name: str | None = None
    alias: str | None = None

    def to_dict(self, camel_case: bool = True) -> dict:
        return {
            **super().to_dict(camel_case),
            'subType': self._TYPE.value,
        }


@dataclass
class MetricTemplateUnit(TemplateUnit):
    """Class representation of single metric element.

    Attributes:
        elements (list[MetricElement]): List containing
            metric elements used in template.
    """

    _TYPE = TemplateUnitType.METRICS
    _FROM_DICT_MAP = {
        'elements': [MetricElement.from_dict],
    }

    elements: list[MetricElement]


@dataclass
class Template(Dictable):
    """Class representation of single template.

    Attributes:
        rows (list[TemplateUnit], optional): List containing
            attribute and metric elements used in template rows.
        columns (list[TemplateUnit], optional): List containing
            attribute and metric elements used in template columns.
        page_by (list[TemplateUnit], optional): List containing
            attribute elements used in template for paging.
    """

    _FROM_DICT_MAP = {
        'rows': [TemplateUnit.dispatch],
        'columns': [TemplateUnit.dispatch],
        'page_by': [TemplateUnit.dispatch],
    }

    rows: list[TemplateUnit] = field(default_factory=list)
    columns: list[TemplateUnit] = field(default_factory=list)
    page_by: list[TemplateUnit] = field(default_factory=list)


class DataLanguageType(AutoName):
    """Enumeration constant indicating type of data language
    configuration type."""

    PROJECT_DEFAULT = auto()
    ALL_PROJECT_DATA_LANGUAGE = auto()
    SPECIFIC_LANGUAGES = auto()


class DataRefreshType(AutoName):
    """Enumeration constant indicating type of data refresh
    operation."""

    REPLACE = auto()
    ADD = auto()
    DYNAMIC_REFRESH = auto()
    UPSERT = auto()


@dataclass
class DataLanguages(Dictable):
    """Class representation of data language configuration.

    Attributes:
        data_language_type (DataLanguageType): Type of used
            data language configuration.
        selected_languages (list[str], optional): List with
            selected languages passed as abbreviations.
    """

    _FROM_DICT_MAP = {
        'data_language_type': DataLanguageType,
    }

    data_language_type: DataLanguageType = DataLanguageType.PROJECT_DEFAULT
    selected_languages: list[str] | None = field(default_factory=list)


@dataclass
class DataPartition(Dictable):
    """Class representation of data partition.

    Attributes:
        partition_attribute (SchemaObjectReference, optional): Reference
            to attribute used in partition.
        number_of_partitions (int, optional): Number of data partitions.
        fetch_data_slices_in_parallel (bool, optional): Bool indicating
            whether to fetch data in parallel, default False.
    """

    partition_attribute: SchemaObjectReference | None = None
    number_of_partitions: int | None = None
    fetch_data_slices_in_parallel: bool = False

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: Optional['Connection'] = None,
        to_snake_case: bool = True,
    ) -> T:
        """Initialize DataPartition object from dictionary."""
        source = camel_to_snake(source)
        obj = super().from_dict(
            source=source, connection=connection, to_snake_case=to_snake_case
        )
        obj.partition_attribute = (
            SchemaObjectReference.from_dict(partition_attribute)
            if (partition_attribute := source.get('partition_attribute'))
            else None
        )
        return obj

    def to_dict(self, camel_case: bool = True) -> dict:
        result = super().to_dict()
        if self.partition_attribute:
            result['partitionAttribute']['subType'] = str.lower(
                self.partition_attribute.sub_type.name
            )
        else:
            result['partitionAttribute'] = {}
        return snake_to_camel(result) if camel_case else result


@dataclass
class CubeOptions(Dictable):
    """Class representation of cube options for cube.

    Attributes:
        data_languages (DataLanguages, optional): Information related
            to data languages configuration.
        data_refresh (DataRefreshType, optional): Type of data refresh
            operation, default 'DataRefreshType.REPLACE'.
        data_partition (DataPartition, optional): Object which
            represents information about data partition.
    """

    _FROM_DICT_MAP = {
        'data_languages': DataLanguages.from_dict,
        'data_refresh': DataRefreshType,
        'data_partition': DataPartition.from_dict,
    }

    data_languages: DataLanguages = field(default_factory=DataLanguages)
    data_refresh: DataRefreshType = DataRefreshType.REPLACE
    data_partition: DataPartition = field(default_factory=DataPartition)


@dataclass
class TimeBasedSettings(Dictable):
    """Class representation of time based settings.

    Attributes:
        timezone (SchemaObjectReference, optional): Reference to timezone
            object or None.
        calendar (SchemaObjectReference, optional): Reference to calendar
            object or None.
        enable_timezone_and_calendar_reporting (bool, optional): Bool
            indicating whether to enable timezone and calendar reporting,
            default True.
    """

    timezone: SchemaObjectReference | None = None
    calendar: SchemaObjectReference | None = None
    enable_timezone_and_calendar_reporting: bool = True

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: Optional['Connection'] = None,
        to_snake_case: bool = True,
    ) -> T:
        """Initialize TimeBasedSettings object from dictionary."""
        source = camel_to_snake(source)
        obj = super().from_dict(
            source=source, connection=connection, to_snake_case=to_snake_case
        )
        obj.timezone = (
            SchemaObjectReference.from_dict(timezone)
            if (timezone := source.get('timezone'))
            else None
        )
        obj.calendar = (
            SchemaObjectReference.from_dict(calendar)
            if (calendar := source.get('calendar'))
            else None
        )
        return obj



"""This is the demo script to show how to manage content groups.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.content_group import (
    ContentGroup,
    list_content_groups,
)
from mstrio.project_objects.dashboard import Dashboard
from mstrio.users_and_groups import User, UserGroup

# Define a variable which can be later used in a script
PROJECT_ID = $project_id  # Insert name of project here

conn = get_connection(workstationData, project_id=PROJECT_ID)

# List content groups with different conditions
# Note: No Content Groups exist in a default environment
list_of_all_cgs = list_content_groups(connection=conn)
print(list_of_all_cgs)
list_of_all_cgs_as_dicts = list_content_groups(connection=conn, to_dictionary=True)
print(list_of_all_cgs_as_dicts)

# Define variables which can be later used in a script
CG_NAME = $cg_name
CG_COLOR = $cg_color
CG_OPACITY = $cg_opacity
CG_EMAIL_ENABLED = $cg_email_enabled == 'True'
CG_RECIPIENT_ID = $cg_recipient_id
CG_USER_RECIPIENT_ID = $cg_user_recipient_id
CG_USER_GROUP_RECIPIENT_ID = $cg_user_group_recipient_id

# Create a Content Group
# Recipients can be provided as a User, UserGroup, or an ID of one of them
# Color has to be provided as a hex value, e.g. '#ff2400'
cg = ContentGroup.create(
    connection=conn,
    name=CG_NAME,
    color=CG_COLOR,
    opacity=CG_OPACITY,
    email_enabled=CG_EMAIL_ENABLED,
    recipients=[
        User(connection=conn, id=CG_USER_RECIPIENT_ID),
        UserGroup(connection=conn, id=CG_USER_GROUP_RECIPIENT_ID),
        CG_RECIPIENT_ID,
    ],
)

# Define variables which can be later used in a script
CG_ID = $cg_id

# Get a Content Group by it's id or name
cg = ContentGroup(connection=conn, id=CG_ID)
cg = ContentGroup(connection=conn, name=CG_NAME)

# Define variables which can be later used in a script
DASHBOARD_ID_1 = $dashboard_id_1
DASHBOARD_ID_2 = $dashboard_id_2

# Add Contents to a Content Group
# Accepted content types are: Dashboard, Document, Report and Agent
cg.update_contents(
    content_to_add=[
        Dashboard(connection=conn, id=DASHBOARD_ID_1),
        Dashboard(connection=conn, id=DASHBOARD_ID_2),
    ]
)
# Remove Contents from a Content Group
cg.update_contents(
    content_to_remove=[
        Dashboard(connection=conn, id=DASHBOARD_ID_1),
    ]
)
# Contents can be added and removed in one call
cg.update_contents(
    content_to_add=[
        Dashboard(connection=conn, id=DASHBOARD_ID_1),
    ],
    content_to_remove=[
        Dashboard(connection=conn, id=DASHBOARD_ID_2),
    ],
)

# Get contents of a Content Group
contents = cg.get_contents(project_ids=[PROJECT_ID])

# Define variables which can be later used in a script
CG_NAME_ALTERED = $cg_name_altered
CG_COLOR_ALTERED = $cg_color_altered
CG_OPACITY_ALTERED = $cg_opacity_altered
CG_EMAIL_ENABLED_ALTERED = $cg_email_enabled_altered == 'True'
CG_RECIPIENT_ID_ALTERED = $cg_recipient_id_altered

# Alter a Content Group
cg.alter(
    name=CG_NAME_ALTERED,
    color=CG_COLOR_ALTERED,
    opacity=CG_OPACITY_ALTERED,
    email_enabled=CG_EMAIL_ENABLED_ALTERED,
    recipients=[CG_RECIPIENT_ID_ALTERED],
)

# Delete a Content Group without prompt
cg.delete(force=True)



import inspect
import logging
import math
import os
import re
import time
import warnings
from base64 import b64encode
from collections.abc import Callable
from copy import deepcopy
from datetime import datetime
from enum import Enum
from functools import reduce, wraps
from json.decoder import JSONDecodeError
from pprint import pformat
from time import sleep
from typing import TYPE_CHECKING, Any, TypeVar

import humps
from pypika import Query
from requests import Response

from mstrio import config
from mstrio.helpers import (
    IServerError,
    MstrTimeoutError,
    PromptedContentError,
    VersionException,
)
from mstrio.types import ObjectSubTypes
from mstrio.utils.dict_filter import filter_list_of_dicts
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.time_helper import (
    DatetimeFormats,
    map_datetime_to_str,
    map_str_to_datetime,
)

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.modeling.expression import Expression
    from mstrio.object_management import SearchPattern  # noqa: F401
    from mstrio.project_objects.datasets import OlapCube, SuperCube
    from mstrio.server import Project
    from mstrio.types import ObjectTypes  # noqa: F401
    from mstrio.users_and_groups import User
    from mstrio.utils.entity import EntityBase

logger = logging.getLogger(__name__)


def deprecation_warning(
    deprecated: str,
    new: str | None,
    version: str,
    module: bool = True,
    change_compatible_immediately=True,
):
    """This function is used to provide a user with a warning, that a given
    functionality is now deprecated, and won't be supported from a given
    version.

    Args:
        deprecated (str): name of a functionality that is deprecated
        new (str): name of a functionality that replaces deprecated one
        version (str): version from which deprecated functionality won't be
            supported
        module (bool, optional): Whether deprecated functionality is a module.
        Defaults to True.
        change_compatible_immediately (bool, optional): Whether the new
        functionality can be used immediately. Defaults to True.
    """

    module = " module" if module else ""
    if change_compatible_immediately:
        msg = (
            f"{deprecated}{module} is deprecated and will not be supported starting "
            f"from mstrio-py {version}."
        )
        if new:
            msg += f" Please use {new} instead."
    else:
        suffix = f" and replaced with {new}" if new else ""
        msg = f"From version {version} {deprecated}{module} will be removed{suffix}."
    exception_handler(msg=msg, exception_type=DeprecationWarning)


def url_check(url):
    """Checks the validity of the url required in the connection object and
    returns a validated url."""
    regex = r'^https?://.+$'
    match = re.search(regex, url)
    api_index = url.find('/api')

    if match is None:
        msg = (
            "Please check the validity of the base_url parameter. Typically of the "
            "form 'https://<<MSTR Domain>>/MicroStrategyLibrary/'"
        )
        raise ValueError(msg)
    if api_index != -1:
        url = url[:api_index]
    if url.endswith('/'):
        url = url[:-1]

    return url


def version_cut(version):
    res = ".".join([str(int(i)) for i in version.split(".")])
    return res[:-2]


def get_args_from_func(func: Callable[[Any], Any]):
    signature = inspect.signature(func)
    return list(signature.parameters.keys())


def get_default_args_from_func(func: Callable[[Any], Any]):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def camel_to_snake(
    response: dict | list,
    whitelist: list[str] = None,
) -> dict | list[dict]:
    """Converts dictionary keys from camelCase to snake_case.
    It works recursively for dicts in dicts."""

    whitelist = whitelist or []

    def convert_dict(source):
        return {
            humps.decamelize(key): (
                value
                if not isinstance(value, dict) or key in whitelist
                else convert_dict(value)
            )
            for key, value in source.items()
        }

    if isinstance(response, list):
        return [convert_dict(source) for source in response if isinstance(source, dict)]
    elif isinstance(response, dict):
        return convert_dict(response)

    raise ValueError("Not supported data type for camel_to_snake conversion")


def snake_to_camel(
    response: dict | list,
    whitelist: list[str] = None,
) -> dict | list[dict]:
    """Converts dictionary keys from snake_case to camelCase.
    It works recursively for dicts in dicts."""
    whitelist = whitelist or []

    def convert_dict(source):
        return {
            humps.camelize(key): (
                value
                if not isinstance(value, dict) or key in whitelist
                else convert_dict(value)
            )
            for key, value in source.items()
        }

    if isinstance(response, list):
        return [convert_dict(source) for source in response if isinstance(source, dict)]
    elif isinstance(response, dict):
        return convert_dict(response)

    raise ValueError("Not supported data type for snake_to_camel conversion")


def check_duplicated_column_names(data_frame):
    list_of_columns = list(data_frame.columns)
    return len(list_of_columns) == len(set(list_of_columns))


def exception_handler(
    msg: str, exception_type: type[Exception] = Exception, stack_lvl: int = 2
):
    """Generic error message handler.

    Args:
        msg (str): Message to print in the Exception or Warning
        exception_type (Exception): Instance of Exception or Warning class
        stack_lvl (int, optional): controls how deep the stacktrace will be
    """
    if not isinstance(exception_type, type) or not issubclass(
        exception_type, Exception
    ):
        raise ValueError("exception_type has to be a subclass of Exception class")
    elif issubclass(exception_type, Exception) and not issubclass(
        exception_type, Warning
    ):
        raise exception_type(msg)
    elif issubclass(exception_type, Warning):
        warnings.warn(msg, exception_type, stacklevel=stack_lvl)


def response_handler(
    response: 'Response',
    msg: str | None = None,
    throw_error: bool = True,
    verbose: bool = True,
    whitelist: list | None = None,
):
    """Generic error message handler for transactions against I-Server.

    Args:
        response: Response object returned by HTTP request.
        msg (str, optional): Message to print in addition to any
            server-generated error message(s).
        throw_error (bool, optional): Flag indicates if the error should
            be thrown (defaults to True).
        verbose (bool, optional): controls if messages/errors will be printed
            (defaults to True).
        whitelist(list, optional): list of tuples of I-Server Error and
            HTTP errors codes respectively (you may add optional third value
            as string, representing message that the error need to contain to
            be whitelisted), which will not be handled (defaults to None).
            i.e. whitelist = [
                ('ERR001', 500),('ERR004', 404),('ERR003', 403, 'not allowed')
            ]
    """
    whitelist = whitelist or []

    try:
        logger.debug(f"{response} url = '{response.url}'")
        logger.debug(f"headers = {pformat(response.headers)}")
        logger.debug(f"content = {response.text}")

        if (
            response.ok
            and response.text == ''
            and (200 < response.status_code < 300 or response.request.method == 'HEAD')
        ):
            # we can expect empty response body and it's fine
            return

        res: dict | list = response.json()

        if response.ok:
            return

        if res.get('errors') is not None:
            res = res.get('errors')
            if len(res) > 0:
                res = res[0]
            else:
                res = {}

        server_code = res.get('code')
        server_msg = res.get('message')
        ticket_id = res.get('ticketId')
        iserver_code = res.get('iServerCode')

        def check_if_whitelisted():
            if not whitelist:
                return False
            for item in whitelist:
                if len(item) == 2:
                    code, http_code = item
                    if (server_code, response.status_code) == (code, http_code):
                        return True
                elif len(item) == 3:
                    code, http_code, part_of_msg = item
                    if (server_code, response.status_code) == (
                        code,
                        http_code,
                    ) and part_of_msg in server_msg:
                        return True
            return False

        if not check_if_whitelisted():
            if (
                server_code == 'ERR004'
                and response.status_code == 404
                and server_msg == 'HTTP 404 Not Found'
            ) or (
                server_code == 'ERR001'
                and response.status_code == 405
                and server_msg == 'HTTP 405 Method Not Allowed'
            ):
                msg = (
                    "This REST API functionality is not supported on this version "
                    f"of the I-Server: {version_cut(config.iserver_version)}."
                )
                exception_handler(msg, exception_type=VersionException)
            elif iserver_code == -2147206497:  # MSI_REQUEST_TIMEOUT on server-side
                raise MstrTimeoutError(res)
            elif iserver_code == -2147468903:
                raise PromptedContentError(
                    'Prompted content subscription creation is not supported.'
                )
            if verbose:
                logger.error(
                    (f'{msg}\n' if msg else '')
                    + f'I-Server Error {server_code}, {server_msg}\n'
                    f'Ticket ID: {ticket_id}'
                )
            if throw_error and server_code:
                raise IServerError(
                    message=(
                        f"{server_msg}; code: '{server_code}', ticket_id: '{ticket_id}'"
                    ),
                    http_code=response.status_code,
                )
    except JSONDecodeError:
        logger.debug(f"Response body: {response.text}")

        if verbose:
            logger.error(
                (f"{msg}\n" if msg else '')
                + "Could not decode the response from the I-Server. Please check if "
                "I-Server is running correctly"
            )
            # raise error if I-Server response cannot be decoded
            if throw_error:
                response.raise_for_status()


def get_response_json(response: 'Response', /, **kwargs) -> dict:
    """Get JSON from response object in a safe, error-handled way.

    Under the hood, it uses `response_handler` to handle errors but returns
    simply `response.json()` if the response is OK.

    Args:
        response: Response object returned by HTTP request.
        **kwargs: Additional arguments to pass to `response_handler`.

    Returns:
        dict: Parsed JSON from the response.
    """
    if not response.ok:
        response_handler(response, **kwargs)
        # if `response_handler` does not raise an error, it means
        # we want to fall through and return the JSON,
        # even if the response is not OK.

    return response.json()


def fallback_on_timeout(min_limit: int = 50):
    """Return a decorator, which decorates a function with one argument,
    `limit`, to retry with half as big limit if it encounters a timeout error.

    Do note that the wrapped function will return a (result, working_limit)
    tuple, not just the result, to allow for using the working limit further.

    Args:
        min_limit: The minimum limit that will be attempted as retry. If the
            wrapper recurses past it, it will just raise the error back.
    """

    def decorate(func: Callable[[int], Any]):
        """Replaces `fot_wrapper`'s __name__, __doc__ and __module__ with those
        of `func`."""

        @wraps(func)
        def fot_wrapper(limit: int) -> tuple[Any, int]:
            try:
                return func(limit), limit
            except MstrTimeoutError as err:
                new_limit = limit // 2
                if new_limit >= min_limit:
                    logger.warning(
                        f"Timeout hit when executing {func.__name__} with limit "
                        f"{limit}, retrying with limit {new_limit}"
                    )
                    return fot_wrapper(new_limit)
                else:
                    raise err

        return fot_wrapper

    return decorate


def get_parallel_number(total_chunks):
    """Returns the optimal number of threads to be used for downloading
    cubes/reports in parallel."""

    threads = min(8, os.cpu_count() + 4)
    if total_chunks > 0:
        threads = min(total_chunks, threads)

    return threads


def _prepare_objects(
    objects: dict | list[dict],
    filters: dict | None = None,
    dict_unpack_value: str | None = None,
    project_id: str | None = None,
):
    if isinstance(objects, dict) and dict_unpack_value:
        objects = objects[dict_unpack_value]
    objects = camel_to_snake(objects)
    if filters:
        objects = filter_list_of_dicts(objects, **filters)  # type: ignore
    if project_id:
        for obj in objects:
            obj.setdefault('project_id', project_id)
    return objects


def get_total_count_of_objects(
    source: Callable[["Connection", int], Response] | Response,
    connection: "Connection | None" = None,
    **kwargs: Any,
) -> int:
    """Get total count of objects from server Response object
    or getter returning server Response object.

    Args:
        source: GET API wrapper function that will return a Response object
            (fn accepting `limit` parameter) or Response object with
            the total count in headers
        connection: Strategy One REST API connection object if `source`
            is a callable.
        **kwargs: additional parameters to pass to the `source` function if it
            is a callable

    Returns:
        int: total count of objects
    """
    # circular import prevention
    from mstrio.connection import Connection
    from mstrio.utils.api_helpers import is_response_like

    HEADER_KEY = 'x-mstr-total-count'

    if is_response_like(source):
        assert HEADER_KEY in source.headers, (
            f"Response headers do not contain '{HEADER_KEY}' entry. "
            "Cannot get total count of objects from this response."
        )
        v = source.headers.get(HEADER_KEY)
        assert isinstance(v, int) or (
            isinstance(v, str) and v.isnumeric()
        ), f"Expected '{HEADER_KEY}' header to be numerical, got {v}."
        return int(v)

    if callable(source):
        kwargs.pop('limit', None)  # limit for count is forced in this fn

        assert isinstance(connection, Connection), (
            "Expected 'connection' to be a Connection object when 'source' "
            "is a callable."
        )
        return get_total_count_of_objects(
            source=source(connection, limit=1, **kwargs),
            connection=connection,
        )

    raise TypeError(
        f"Expected 'source' to be a Response object or a callable, got {type(source)}"
    )


def fetch_objects_async(
    connection: "Connection",
    api: Callable,
    async_api: Callable,
    limit: int | None,
    chunk_size: int,
    filters: dict,
    error_msg: str | None = None,
    dict_unpack_value: str | None = None,
    **kwargs,
) -> list:
    """Get all objects asynchronously. Optionally filter the objects using
    `filters` parameter. Works only for endpoints with `limit` and `offset`
    query parameter (pagination).

    Args:
        connection: Strategy One REST API connection object
        api: GET API wrapper function that will return list of objects in bulk
        async_api: asynchronous wrapper of the `api` function
        dict_unpack_value: if the response needs to be unpacked to get into
            the values, specify the keyword
        limit: cut-off value for the number of objects returned
        chunk_size: number of objects in each chunk
        error_msg: specifies error_msg for failed requests
        **filters: dict that specifies filter expressions by which objects will
            be filtered locally
        kwargs: all specific parameters that the api methods require that need
            to be additionally specified
    """
    validate_param_value('limit', limit, int, min_val=1, special_values=[None])
    offset = 0
    chunk_size = min(limit, chunk_size) if limit else chunk_size
    all_objects = []

    # Extract parameters of the api wrapper and set them using the kwargs
    args = get_args_from_func(api)
    param_value_dict = {
        arg: kwargs.get(arg)
        for arg in args
        if arg not in ['connection', 'limit', 'offset', 'error_msg']
    }
    response = api(
        connection=connection,
        offset=offset,
        limit=chunk_size,
        error_msg=error_msg,
        **param_value_dict,
    )
    project_id = kwargs.get('project_id') or kwargs.get('project')
    objects = _prepare_objects(response.json(), filters, dict_unpack_value, project_id)
    all_objects.extend(objects)
    current_count = offset + chunk_size
    total_objects = get_total_count_of_objects(response)
    total_objects = min(limit, total_objects) if limit else total_objects

    if total_objects > current_count:
        it_total = math.ceil(total_objects / chunk_size)
        threads = get_parallel_number(it_total)
        with FuturesSessionWithRenewal(
            connection=connection, max_workers=threads
        ) as session:
            # Extract parameters of the api wrapper and set them using kwargs
            param_value_dict = auto_match_args(
                api,
                kwargs,
                exclude=[
                    'connection',
                    'limit',
                    'offset',
                    'future_session',
                    'error_msg',
                ],
            )
            futures = [
                async_api(
                    future_session=session,
                    offset=offset,
                    limit=chunk_size,
                    **param_value_dict,
                )
                for offset in range(current_count, total_objects, chunk_size)
            ]

        for f in futures:
            response = f.result()
            if not response.ok:
                response_handler(response, error_msg, throw_error=False)
                continue
            objects = _prepare_objects(
                response.json(), filters, dict_unpack_value, project_id
            )
            all_objects.extend(objects)
    return all_objects


def fetch_objects(
    connection: "Connection",
    api: Callable,
    limit: int | None,
    filters: dict,
    error_msg: str | None = None,
    dict_unpack_value: str | None = None,
    **kwargs,
) -> list:
    """Fetch and prepare objects. Optionally filter the objects by using the
    filters parameter. This function only supports endpoints without pagination.

    Args:
        connection: Strategy One REST API connection object
        api: GET API wrapper function that will return list of objects in bulk
        dict_unpack_value: if the response needs to be unpacked to get into
            the values specify the keyword
        limit: value to be provided in REST request to limit count of
            returned objects
        error_msg: specifies error_msg for failed requests
        **filters: dict that specifies filter expressions by which objects will
            be filtered locally
        kwargs: all specific parameters that the api methods require that need
            to be additionally specified
    """
    validate_param_value('limit', limit, int, min_val=1, special_values=[None])

    # Extract parameters of the api wrapper and set them using kwargs
    dic = {**kwargs}
    if limit is not None:
        dic['limit'] = limit  # yes, overwrite if exists

    param_value_dict = auto_match_args(api, dic, exclude=['connection', 'error_msg'])
    response = api(connection=connection, error_msg=error_msg, **param_value_dict)

    if response.ok:
        project_id = kwargs.get('project_id') or kwargs.get('project')
        objects = _prepare_objects(
            response.json(), filters, dict_unpack_value, project_id
        )
        if limit is not None:
            # If response returned more than `limit` objects -> cut-off.
            # This can happen if limiting is done ex. per node.
            # Some endpoints return `limit` objects PER NODE.
            # We don't validate: just get first `limit` amount of them.
            objects = objects[:limit]
        return objects
    else:
        return []


def key_fn_for_sort_object_properties(source: Any) -> int:
    """Sort all properties of an object representing an MSTR object."""
    preferred_order = {
        'id': 1,
        'name': 2,
        'description': 3,
        'alias': 4,
        'type': 5,
        'subtype': 6,
        'ext_type': 7,
        'acg': 101,
        'acl': 102,
    }
    return preferred_order.get(source, 50)


def auto_match_args(
    func: Callable,
    param_dict: dict,
    exclude: list | None = None,
    include_defaults: bool = True,
    id_weak_match: bool = False,
) -> dict:
    """Automatically match dict data to function arguments.

    Handles default parameters. Extracts value from Enums. Returns matched
    arguments as dict.

    Note:
        don't use it for alter purposes as, changing parameter value
        back to default currently doesn't work (Rework line 413?)

    Args:
        func: function for which args will be matched
        param_dict: dict to use for matching the function args
        exclude: set `exclude` parameter to exclude specific param-value pairs
        include_defaults: if `False` then values which have the same value as
            default will not be included in the result
        id_weak_match: if `True`, the function will try to match IDs even if
            they are not exact (e.g. by ignoring certain prefixes)
    Raises:
        KeyError: could not match all required arguments
    """

    exclude = exclude or []
    args = get_args_from_func(func)
    default_dict = get_default_args_from_func(func)

    param_value_dict = {}
    for arg in args:
        if arg in exclude:
            continue

        val = get_val_safely(param_dict, arg, default_dict)
        if not include_defaults and arg in default_dict and val == default_dict[arg]:
            continue

        if (
            id_weak_match
            and val is None
            and arg != 'project_id'
            and arg.endswith('_id')
        ):
            val = get_val_safely(param_dict, 'id', default_dict)

        val = val.value if isinstance(val, Enum) else val
        param_value_dict.update({arg: val})

    return param_value_dict


def get_val_safely(param_dict: dict, key: str, default_dict: dict) -> Any:
    """Safely get a value from a dictionary."""
    return (
        param_dict.get(key)
        if param_dict.get(key) is not None
        else default_dict.get(key)
    )


def flatten2list(object) -> list:
    """Flatten to list nested objects of type list, tuple, sets."""
    gather = []
    for item in object:
        if isinstance(item, (list, tuple, set)):
            gather.extend(flatten2list(item))
        else:
            gather.append(item)
    return gather


def dict_compare(d1, d2):
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    added = d1_keys - d2_keys
    removed = d2_keys - d1_keys
    modified = {o: (d1[o], d2[o]) for o in intersect_keys if d1[o] != d2[o]}
    same = {o for o in intersect_keys if d1[o] == d2[o]}
    return added, removed, modified, same


def __validate_single_param_value(
    value,
    param_name,
    data_type,
    max_val,
    min_val,
    regex,
    valid_example,
    inv_val,
    special_values=None,
):
    special_values = special_values or []
    if value in special_values:
        return True

    if max_val is not None and value > max_val:
        msg = f"'{param_name}' has to be less than or equal to {max_val}"
        exception_handler(msg, inv_val)
        return False
    elif min_val is not None and value < min_val:
        msg = f"'{param_name}' has to be greater than or equal to {min_val}"
        exception_handler(msg, inv_val)
        return False
    elif regex is not None and not re.match(regex, value):
        pattern = valid_example or regex
        msg = f"'{param_name}' has to match pattern '{pattern}'"
        exception_handler(msg, inv_val)
        return False
    elif (
        all(cond is None for cond in [max_val, min_val, regex])
        and special_values
        and str not in data_type
    ):
        msg = f"'{param_name}' has to be one of {special_values}"
        exception_handler(msg, inv_val)
        return False
    return True


def validate_param_value(
    param_name,
    param_val,
    data_type,
    max_val=None,
    min_val=None,
    special_values=None,
    regex=None,
    exception=True,
    valid_example=None,
) -> bool:
    """Validate param data type and optionally max, min special values.

    Raise:
        TypeError
        ValueError
    """
    special_values = special_values or []
    inv_type = TypeError if exception else Warning
    inv_val = ValueError if exception else Warning
    data_type = data_type if isinstance(data_type, list) else [data_type]

    if any(x == param_val and isinstance(x, type(param_val)) for x in special_values):
        return True

    if type(param_val) not in data_type:
        msg = f"'{param_name}' needs to be of type {data_type}"
        exception_handler(msg, inv_type)
        return False
    if isinstance(param_val, list):
        return all(
            __validate_single_param_value(
                value,
                param_name,
                data_type,
                max_val,
                min_val,
                regex,
                valid_example,
                inv_val,
                special_values,
            )
            for value in param_val
        )

    return __validate_single_param_value(
        param_val,
        param_name,
        data_type,
        max_val,
        min_val,
        regex,
        valid_example,
        inv_val,
        special_values,
    )


def extract_all_dict_values(list_of_dicts: list[dict]) -> list[Any]:
    """Extract list of dicts values into list."""
    all_options = []
    for option in list_of_dicts:
        all_options.extend(option.values())
    return all_options


def delete_none_values(
    source: dict, *, whitelist_attributes: list | None = None, recursion: bool
) -> dict:
    """Delete keys with None values from dictionary.

    Args:
        source (dict): dict object from which none values will be deleted
        whitelist_attributes (list): keyword-only argument containing name
            of attribute, which should be left alone even if they contain
            none values
        recursion (bool): flag that turns recursion on or off
    """
    whitelist = whitelist_attributes or []

    new_dict = {}
    for key, value in source.items():
        if recursion and isinstance(value, dict):
            new_dict[key] = delete_none_values(
                value, whitelist_attributes=whitelist, recursion=recursion
            )
        elif value not in [[], {}, None] or key in whitelist:
            new_dict[key] = value

    return new_dict


def get_objects_id(obj, obj_class):
    if isinstance(obj, str):
        return obj
    elif isinstance(obj, obj_class):
        return obj.id
    return None


def merge_id_and_type(
    object_id: str,
    object_type: 'ObjectTypes | ObjectSubTypes | int',
    error_msg: str | None = None,
) -> str:
    if not object_id or not object_type:
        exception_handler(
            msg=error_msg or "Please provide both `id` and `type`.",
            exception_type=AttributeError,
        )
    object_id = get_objects_id(object_id, type(object_id))
    object_type = (
        get_enum_val(object_type, type(object_type))
        if isinstance(object_type, Enum)
        else object_type
    )
    return f'{object_id};{object_type}'


def rsetattr(obj, attr, val):
    """Recursive setattr. This can only modify last attr in the chain.
    For example rsetattr(obj, 'attr1.attr2.attr3', new_value) won't work
    if attr2 doesn't exist.

    Args:
        obj: An object that is edited
        attr: attribute name chain/path
        val: new value
    Returns:
        None on success
    Raises:
        AttributeError on failure, when the 'attr' is incorrect
    Example:
        rsetattr(obj, 'attr1.attr2', new_value)
    """

    pre, _, post = attr.rpartition('.')
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *default):
    """Recursive getattr. Third parameter is the default value.

    Args:
        obj: An object that is inspected
        attr: attribute name chain/path
        default (optional): value returned if 'attr' not found
    Returns:
        value of 'attr' on success. On failure, default value.
    Raises:
        AttributeError on failure, when the 'attr' is incorrect
        and no default provided
    Example:
        rgetattr(obj, 'attr1.attr2', 'default value')
    """

    def _getattr(obj, attr):
        return getattr(obj, attr, *default)

    return reduce(_getattr, [obj] + attr.split('.'))


def filter_params_for_func(
    func: Callable, params: dict, exclude: list | None = None
) -> dict:
    """Filter dict of parameters and return only those that are parameters
    of a `func`.
    Mainly used in `EntityBase.alter()`, before calling
    `EntityBase._alter_properties()` as shown in `Example`.

    Args:
        func: a function that the params are going to be fit for
        params: a dict with parameters that will be filtered
        exclude: set `exclude` parameter to exclude specific param-value pairs
    Returns:
        a dict with parameters of `func` that exist in `params` dict
    Example:
        params = filter_params_for_func(self.alter, locals())
        self._alter_properties(**params)
    """
    args = get_args_from_func(func)
    defaults_dict = get_default_args_from_func(func)
    exclude = exclude or []
    properties = {}

    for arg in args:
        if arg in exclude:
            continue
        elif params.get(arg) is not None:
            properties.update({arg: params.get(arg)})
        elif defaults_dict.get(arg) is not None:
            properties.update({arg: defaults_dict.get(arg)})
    return properties


T = TypeVar("T")


def filter_obj_list(obj_list: list[T], **filters: dict[str, Any]) -> list[T]:
    """
    Filter a list of objects by providing one or more key-value pair filters.
    """
    return [
        obj for obj in obj_list if all(getattr(obj, f) == v for f, v in filters.items())
    ]


def choose_cube(
    connection: 'Connection', cube_dict: dict
) -> 'OlapCube | SuperCube | None':
    """Return correct cube object based on dictionary with cube's info.

    Note: In case of wrong subtype, `None` is returned.
    """
    cube_subtype = cube_dict['subtype']
    if cube_subtype == ObjectSubTypes.OLAP_CUBE.value:
        from mstrio.project_objects.datasets.olap_cube import OlapCube

        return OlapCube.from_dict(cube_dict, connection)
    elif cube_subtype == ObjectSubTypes.SUPER_CUBE.value:
        from mstrio.project_objects.datasets.super_cube import SuperCube

        return SuperCube.from_dict(cube_dict, connection)


def get_temp_connection(
    connection: "Connection",
    project_id: str | None = None,
) -> "Connection":
    """Return a temporary connection object with a selected project.

    Args:
        connection: Strategy One connection object
        project_id: Project ID
    """
    if project_id:
        temp_conn = deepcopy(connection)
        temp_conn.select_project(project_id=project_id)
    else:
        temp_conn = connection
    return temp_conn


class Dictable:
    """The fundamental class in mstrio-py package. Includes support for
    converting an object to a dictionary, and creating an object from a
    dictionary."""

    _FROM_DICT_MAP: dict[str, Callable] = {}  # map attributes to Enums and components
    # list of attribute name, which are allowed to have none values
    # in dict returned by .to_dict()
    _ALLOW_NONE_ATTRIBUTES: list[str] = []
    _KEEP_CAMEL_CASE: list[str] = []

    @classmethod
    def _unpack_objects(cls, key, val, camel_case=True):
        """Unpack Enums, Dictable obj, list of Enums"""
        if isinstance(val, datetime):
            return map_datetime_to_str(key, val, cls._FROM_DICT_MAP)
        elif isinstance(val, Enum):
            return val.value
        elif isinstance(val, Dictable):
            return val.to_dict(camel_case)
        elif isinstance(val, list):
            return [cls._unpack_objects(key, v, camel_case) for v in val]
        return val

    @classmethod
    def _dict_to_obj(cls, connection, val, key):
        def constructor():
            if isinstance(cls._FROM_DICT_MAP[key], DatetimeFormats):
                return map_str_to_datetime(key, val, cls._FROM_DICT_MAP)
            elif isinstance(cls._FROM_DICT_MAP[key], type(Enum)):
                return cls._FROM_DICT_MAP[key](val)
            elif isinstance(cls._FROM_DICT_MAP[key], type(Dictable)):
                return cls._FROM_DICT_MAP[key].from_dict(val)
            elif isinstance(cls._FROM_DICT_MAP[key], list):
                if isinstance(cls._FROM_DICT_MAP[key][0], list):
                    # for: List[List[handling_cls]]
                    handling_cls = cls._FROM_DICT_MAP[key][0][0]
                    return [[handling_cls.from_dict(item) for item in v] for v in val]
                elif (
                    all(isinstance(v, type(Enum)) for v in cls._FROM_DICT_MAP[key])
                    and val is not None
                ):
                    return [cls._FROM_DICT_MAP[key][0](v) for v in val]
                elif all(
                    isinstance(v, type(Dictable)) for v in cls._FROM_DICT_MAP[key]
                ):
                    return [cls._FROM_DICT_MAP[key][0].from_dict(v) for v in val]
                elif callable(cls._FROM_DICT_MAP[key][0]):
                    return [cls._FROM_DICT_MAP[key][0](v, connection) for v in val]
            else:
                return cls._FROM_DICT_MAP[key](val, connection)

        return val if key not in cls._FROM_DICT_MAP else constructor()

    def to_dict(self, camel_case: bool = True) -> dict:
        """Converts an object to a dictionary excluding object's private
        properties. When converting the object to a dictionary, the object's
        attributes become the dictionary's keys and are in camel case by default
        Attribute values stored as objects are automatically converted to
        non-/ primitive data structures.

        Args:
            camel_case (bool, optional): Set to True if attribute names should
                be converted from snake case to camel case. Defaults to True.

        Returns:
            dict: A dictionary representation of object's attributes and values.
                By default, the dictionary keys are in camel case.
        """

        hidden_keys = [
            '_fetched_attributes',
            '_altered_properties',
            '_connection',
            'connection',
            '_type',
            '_WITH_MISSING_VALUE',
        ]
        cleaned_dict = self.__dict__.copy()

        properties = get_object_properties(self)
        for prop in properties:
            to_be_deleted = '_' + prop
            cleaned_dict[prop] = cleaned_dict.pop(to_be_deleted, None)
        result = {
            key: self._unpack_objects(key, val, camel_case)
            for key, val in cleaned_dict.items()
            if key not in hidden_keys
        }

        result = delete_none_values(
            result, whitelist_attributes=self._ALLOW_NONE_ATTRIBUTES, recursion=False
        )
        result = {
            key: result[key]
            for key in sorted(result, key=key_fn_for_sort_object_properties)
        }
        return (
            snake_to_camel(result, whitelist=self._KEEP_CAMEL_CASE)
            if camel_case
            else result
        )

    @classmethod
    def from_dict(
        cls: T,
        source: dict[str, Any],
        connection: 'Connection | None' = None,
        to_snake_case: bool = True,
        with_missing_value: bool = False,
    ) -> T:
        """Creates an object from a dictionary. The dictionary's keys in camel
        case are changed to object's attribute names (by default in snake case)
        and dict values are composed to their proper data types such as Enums,
        list of Enums etc. as specified in _FROM_DICT_MAP.

        Args:
            cls (T): Class (type) of an object that should be created.
            source (Dict[str, Any]): A dictionary from which an object will be
                constructed.
            connection (Connection, optional): A MSTR Connection object.
                Defaults to None.
            to_snake_case (bool, optional): Set to True if attribute names
                should be converted from camel case to snake case. Defaults to
                True.
            with_missing_value: (bool, optional): If True, class attributes
                possible to fetch and missing in `source` will be set as
                `MissingValue` objects.

        Returns:
            T: An object of type T.
        """
        object_source = (
            camel_to_snake(source, whitelist=cls._KEEP_CAMEL_CASE)
            if to_snake_case
            else source
        )

        if connection is not None:
            object_source["connection"] = connection

        args = {
            key: cls._dict_to_obj(connection, val, key)
            for key, val in object_source.items()
            if key in cls.__init__.__code__.co_varnames
        }
        obj = cls(**args)  # type: ignore
        return obj

    @classmethod
    def bulk_from_dict(
        cls: T,
        source_list: list[dict[str, Any]],
        connection: 'Connection | None' = None,
        to_snake_case: bool = True,
        with_missing_value: bool = False,
    ) -> list[T]:
        """Creates multiple objects from a list of dictionaries. For each
        dictionary provided the keys in camel case are changed to object's
        attribute names (by default in snake case) and dict values are composed
        to their proper data types such as Enums, list of Enums etc. as
        specified in the object's _FROM_DICT_MAP.

        Args:
            cls (T): Class (type) of the objects that should be created.
            source_list (List[Dict[str, Any]]): A list of dictionaries from
                which the objects will be constructed.
            connection (Connection, optional): A MSTR Connection object.
                Defaults to None.
            to_snake_case (bool, optional): Set to True if attribute names
                should be converted from camel case to snake case. Defaults to
                True.
            with_missing_value: (bool, optional): If True, class attributes
                possible to fetch and missing in `source` will be set as
                `MissingValue` objects.

        Returns:
            T: A list of objects of type T.
        """
        return [
            cls.from_dict(
                source=source,
                connection=connection,
                to_snake_case=to_snake_case,
                with_missing_value=with_missing_value,
            )
            for source in source_list
        ]

    def __repr__(self) -> str:
        from mstrio.utils.entity import auto_match_args_entity

        param_dict = auto_match_args_entity(
            self.__init__,
            self,
            exclude=['self'],
            include_defaults=False,
        )
        formatted_params = ', '.join(
            (f'{param}={repr(value)}' for param, value in param_dict.items())
        )
        return f'{self.__class__.__name__}({formatted_params})'


def is_dashboard(view_media: int) -> bool:
    """Documents and dashboards have the same type and subtype when returned
    from search api. They can be distinguished only by view_media value.
    """
    # bits 30-29 are 1, bits 31 and 27 are 0 (011x 0xxx, then 3 bytes)
    return view_media & 0xE800_0000 == 0x6000_0000


def is_document(view_media: int) -> bool:
    """Documents and dashboards have the same type and subtype when returned
    from search api. They can be distinguished only by view_media value.
    """
    return not is_dashboard(view_media)


def rename_dict_keys(source: dict, mapping: dict) -> dict:
    """Rename dict keys according to mapping.

    Args:
        source (dict): An original dictionary, which keys are to be renamed.
        mapping (dict): A dictionary containing mapping of keys.

    Returns:
        dict: A dictionary with keys renamed.
    """
    for rest_name, python_name in mapping.items():
        if rest_name in source:
            old = source.pop(rest_name)
            if python_name:
                source[python_name] = old
    return source


def verify_project_status(
    project: 'Project',
    correct_statuses: list[str] | str,
    node: str | None = None,
    timeout: int = 60,
) -> bool:
    """Verify if provided status is correct for given project.

    Args:
        project (Project): Project for which statuses will be verified.
        correct_statuses (list[str], str): A list of correct statuses
            or just one status.
        node (str, optional): Node name on which status should be verified,
            represented as name of first node if not provided.

    Returns:
        bool: True if status is correct and False otherwise.
    """

    def get_status(project: 'Project', node: str | None = None) -> str:
        node_name = node if node else project.nodes[0]['name']
        nodes_filtered = [node for node in project.nodes if node['name'] == node_name]

        if not nodes_filtered:
            raise ValueError(
                f"Node {node} not found. Available nodes "
                f"{[node['name'] for node in project.nodes]}"
            )

        projects_filtered = [
            proj
            for node in nodes_filtered
            for proj in node['projects']
            if proj['id'] == project.id
        ]

        return projects_filtered[0]['status']

    status = get_status(project=project, node=node)
    iteration = 0
    correct_statuses = (
        correct_statuses if isinstance(correct_statuses, list) else [correct_statuses]
    )

    while status not in correct_statuses and iteration < timeout:
        time.sleep(1)
        project.fetch('nodes')
        status = get_status(project=project, node=node)
        iteration += 1

    return status in correct_statuses


def find_object_with_name(
    connection: 'Connection',
    cls,
    name: str,
    listing_function: callable,
    search_pattern: 'SearchPattern | None' = None,
) -> dict:
    """Find objects with given name if no id is given.

    Args:
        connection: A Strategy One connection object
        name: name of the object. Defaults to None.
        listing_function: function called to list all the objects
            with given name
        search_pattern: search pattern used to find the object.

    Returns:
        dict: object properties in a dictionary.

    Raises:
        ValueError: if both `id` and `name` are not provided,
            if there is more than 1 object with the given `name` or
            if object with the given `name` doesn't exist.
    """
    if search_pattern:
        results = listing_function(
            connection=connection, name=name, search_pattern=search_pattern
        )
    else:
        results = listing_function(connection=connection, name=name)

    if results:
        number_of_objects = len(results)

        if 1 < number_of_objects <= 5:
            error_string = (
                f"There are {number_of_objects} {cls.__name__} objects"
                f" with name: {name}:\n"
            )
            for obj in results:
                path = "/".join(ancestor['name'] for ancestor in obj.ancestors)
                error_string += f"ID: {obj.id} in folder: {path}\n"

            error_string += "Please initialize with ID."

            raise ValueError(error_string)

        if number_of_objects > 5:
            raise ValueError(
                f"There are {number_of_objects} {cls.__name__}"
                " objects with this name. Please initialize with ID."
            )

        return results[0].to_dict()
    else:
        raise ValueError(f"There is no {cls.__name__} with the given name: '{name}'")


def get_object_properties(obj: object) -> set[str]:
    """Extract private object properties.

    Args:
        obj (object): An object which properties will be extracted.

    Returns:
        set[str]: Names of object properties in a set.
    """

    return {
        elem[0]
        for elem in inspect.getmembers(obj.__class__, lambda x: isinstance(x, property))
    }


def encode_as_b64(query: str | Query) -> str:
    """Encodes a query as base64.

    Args:
        query (str | Query): query to be encoded

    Returns:
        Base64 format encoded query."""
    return b64encode(str(query).encode('utf-8')).decode('utf-8')


def get_string_exp_body(expression: str) -> dict:
    return {
        "tokens": [
            {"value": "%", "type": "character"},
            {"value": expression},
        ]
    }


def construct_expression_body(expression: 'str | Expression | dict') -> dict:
    from mstrio.modeling.expression import Expression

    if isinstance(expression, Expression):
        return expression.to_dict()
    if isinstance(expression, str):
        return get_string_exp_body(expression)
    return expression


def deduplicated_name(name: str, existing_names: list[str]) -> str:
    """
    Deduplicate a name by adding a number in parentheses if it already exists.

    Args:
        name: The user-provided name to deduplicate.
        existing_names: A list of existing names to check against.

    Returns:
        The deduplicated name, e.g. 'Schedule (1)'.
    """
    if name not in existing_names:
        return name

    # found duplicate, candidate would be '{name} (1)'
    # find more names that would conflict with the new name

    # trim potential conflicts, e.g. 'Schedule (1)' -> '1'
    trim_index = len(name) + 2
    conflict_strs = [n[trim_index:-1] for n in existing_names if n.startswith(name)]
    conflict_inds: list[int] = [int(n) for n in conflict_strs if n.isnumeric()]

    # find lowest positive integer not in found indices
    i = 1
    while i in conflict_inds:
        i += 1

    return f"{name} ({i})"


def wait_for_stable_status(
    obj: 'EntityBase',
    property: str,
    not_stable_val: list,
    timeout: int = 240,
    interval: int = 1,
) -> bool:
    """Wait for a specific property of an object to reach a stable state.

    Args:
        obj (EntityBase): The object to monitor.
        property (str): The property to check.
        not_stable_val (list): List of values that indicate the property is not
            stable.
        timeout (int, optional): Maximum time to wait in seconds.
            Defaults to 240.
        interval (int, optional): Time between checks in seconds.
            Defaults to 1.

    Returns:
        bool: True if stable state is reached False otherwise.
    """
    start_time = datetime.now()
    while (getattr(obj, property) in not_stable_val) and (
        (datetime.now() - start_time).total_seconds() < timeout
    ):
        obj.fetch()
        sleep(interval)

    return getattr(obj, property) not in not_stable_val


def get_owner_id(
    connection: 'Connection',
    owner: 'str | User | dict | None' = None,
    owner_id: str | None = None,
    owner_username: str | None = None,
) -> str | None:
    """Get owner ID based on provided parameters.

    Args:
        connection (Connection): Strategy One connection object.
        owner (str | User | dict | None): Owner as ID, User object or dict.
            Will take precedence over other parameters.
        owner_id (str | None): Owner's ID.
        owner_username (str | None): Owner's username.

    Returns:
        str | None: User ID of the owner or None if not found.
    """
    from mstrio.users_and_groups import User

    # If owner is a User object, return its ID directly
    if isinstance(owner, User):
        return owner.id

    owner_name = None
    # Dict can contain 'id', but also only 'name'
    if isinstance(owner, dict):
        owner_id = owner.get('id')
        owner_name = owner.get('name')
        owner = None

    # Determine the user identifier to look up
    user_identifier = owner or owner_id or owner_username
    if not user_identifier and not owner_name:
        return None

    # Get user and return ID if found
    if user := get_user_based_on_id_or_username(
        connection, user_identifier, user_identifier, owner_name
    ):
        return user.id

    return None


def get_user_based_on_id_or_username(
    connection: 'Connection',
    user_id: str | None = None,
    user_username: str | None = None,
    user_name: str | None = None,
) -> 'User | None':
    """Get User object based on provided user ID or username.

    Args:
        connection (Connection): Strategy One connection object.
        user_id (str | None): User's ID.
        user_username (str | None): User's username (e.g. 'mstr').
        user_name (str | None): User's name (e.g. 'MSTR User').

    Returns:
        User | None: User object if found, None otherwise.
    """
    from mstrio.users_and_groups import User

    both_the_same = user_id == user_username
    if user_id:
        try:
            with config.temp_verbose_disable():
                return User(connection=connection, id=user_id)
        except IServerError:
            if not both_the_same:
                logger.warning(
                    f"Could not find user with ID '{user_id}'. "
                    "Please provide a valid user ID."
                )
    if user_username:
        try:
            with config.temp_verbose_disable():
                return User(connection=connection, username=user_username)
        except ValueError:
            if not both_the_same:
                logger.warning(
                    f"Could not find user with username '{user_username}'. "
                    "Please provide a valid user username."
                )
    if both_the_same:
        logger.warning(
            f"Could not find user with ID or username '{user_id}'. "
            "Please provide a valid user ID or username."
        )
    if user_name:
        try:
            return User(connection=connection, name=user_name)
        except ValueError:
            if not both_the_same:
                logger.warning(
                    f"Could not find user with name '{user_name}'. "
                    "Please provide a valid user name."
                )
    return None


def is_valid_str_id(str_id: str | None) -> bool:
    """Check if the provided Strategy ID has valid format.

    Args:
        str_id (str | None): The Strategy ID to check.

    Returns:
        bool: True if the ID is valid, False otherwise.
    """
    if not str_id or not isinstance(str_id, str) or len(str_id) != 32:
        return False

    HEX32 = re.compile(r"^[0-9A-F]{32}$")
    return HEX32.match(str_id) is not None



from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(
    err_msg="Error obtaining the list of registered nodes from the Strategy One "
    "deployment."
)
def get_nodes(connection, error_msg=None):
    """Obtain the list of registered nodes from the Strategy One deployment.

    Args:
        connection(object): Strategy One connection object returned by
            'connection.Connection().
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.get(endpoint='/api/registrations/nodes')


@ErrorHandler(
    err_msg="Error obtaining the list of registered services available from "
    "the Strategy One deployment"
)
def get_services(connection, error_msg=None):
    """Obtain the list of registered services available from the Strategy One
    deployment.

    Args:
        connection(object): Strategy One connection object returned by
            'connection.Connection().
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.get(endpoint='/api/registrations/services')


@ErrorHandler(
    err_msg="Error obtaining the metadata information for the registered services"
    " available from the Strategy One deployment."
)
def get_services_metadata(connection, error_msg=None):
    """Obtain the metadata information for the registered services available
    from the Strategy One deployment.

    Args:
        connection(object): Strategy One connection object returned by
            "connection.Connection().
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.get(endpoint='/api/registrations/services/metadata')


@ErrorHandler(err_msg="Error to start/stop service")
def start_stop_service(
    connection, login, password, name, id, address, action='START', error_msg=None
):
    """Start or stop registered service.

    Args:
        connection(object): Strategy One connection object returned by
            'connection.Connection()
        login (string): login for SSH operation
        password (string): password for SSH operation
        name(string): name of the service
        id(string): name of the service
        action(string): one of "START" or "STOP"
        error_msg (string, optional): Custom Error Message for Error Handling
    Returns:
        Complete HTTP response object.
    """

    body = {
        "name": name,
        "id": id,
        "action": action,
        "address": address,
        "login": login,
        "password": password,
    }
    endpoint = '/api/registrations/services/control'
    return connection.post(endpoint=endpoint, json=body)



import logging
from collections.abc import Callable
from typing import Any

from mstrio import config
from mstrio.api import datasources
from mstrio.connection import Connection
from mstrio.datasources import DatasourceConnection, DatasourceInstance, DatasourceLogin
from mstrio.server.language import Language, list_languages
from mstrio.server.project import Project
from mstrio.users_and_groups import User, UserGroup
from mstrio.utils import helper
from mstrio.utils.entity import DeleteMixin, EntityBase
from mstrio.utils.helper import delete_none_values, get_objects_id
from mstrio.utils.resolvers import get_project_id_from_params_set
from mstrio.utils.response_processors import datasources as datasources_processors
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0000')
def list_datasource_mappings(
    connection: 'Connection',
    project: 'Project | str | None' = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    user: User | UserGroup | str | None = None,
    ds_connection: DatasourceConnection | str | None = None,
    datasource: DatasourceInstance | str | None = None,
    login: DatasourceLogin | str | None = None,
    locale: Language | str | None = None,
    default_connection_map: bool = False,
) -> list['DatasourceMap'] | list[dict]:
    """Get list of connection mappings: objects or dicts.
    Optionally filter by specifying filters.
    When returned as dictionaries, connection mappings that have default locale
        will have 'locale' property with empty name and id.

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        default_connection_map: True if requesting default connection mappings.
            Default False
        project: The project (or its id) which mappings are to be fetched.
            Optional unless requesting the default mappings.
        to_dictionary: If True returns dict, by default (False) returns
            DatasourceMap objects.
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        locale: filter list by locale
        login: filter list by datasource login
        datasource: filter list by datasource instance
        ds_connection: filter list by datasource connection
        user: filter list by user or user group

    Returns:
        list['DatasourceMap'] by default
        list[dict] if `to_dictionary` is True

    Examples:
        >>> list_datasource_mappings(connection)
    """
    if default_connection_map:
        project = get_project_id_from_params_set(
            connection,
            project,
            None,
            None,
        )

    return DatasourceMap._list(
        connection=connection,
        project=project,
        to_dictionary=to_dictionary,
        limit=limit,
        user=user,
        ds_connection=ds_connection,
        datasource=datasource,
        login=login,
        locale=locale,
        default_connection_map=default_connection_map,
    )


class DatasourceMap(EntityBase, DeleteMixin):
    """Object representation of Strategy One Connection Mapping

    The connection mapping provides mapping between a user or a user group
    and a datasource login for the given datasource and datasource connection.

    Attributes:
        connection: A Strategy One connection object
        id: ID of connection mapping.
        project: The project the mapping is assigned to.
        default_connection_map: Whether the mapping is the default
            for the project.
        ds_connection: The mapped Datasource Connection
        datasource: The mapped Datasource Instance
        user: The mapped User or UserGroup
        login: The mapped Datasource Login
        locale: The mapping's locale. When empty, it's equivalent to 'default'
            in Workstation and means that default locale for the environment
            will be used.
    """

    _API_GETTERS: dict[str | tuple, Callable] = {
        (
            'id',
            'projectId',
            'connection',
            'datasource',
            'user',
            'login',
            'locale',
        ): datasources.get_datasource_mapping,
    }
    _API_PATCH: list[Callable] = []
    _API_DELETE = staticmethod(datasources.delete_datasource_mapping)
    _FROM_DICT_MAP = {
        'ds_connection': DatasourceConnection.from_dict,
        'datasource': DatasourceInstance.from_dict,
        'user': lambda source, connection: (
            User.from_dict(source, connection)
            if User.from_dict(source, connection).subtype == 8704
            else UserGroup.from_dict(source, connection)
        ),
        'login': DatasourceLogin.from_dict,
        'locale': Language.from_dict,
        'project': lambda source, connection: Project.from_dict(
            {'id': source}, connection
        ),
    }
    _REST_ATTR_MAP = {
        # 'connection': 'ds_connection',
        'project_id': 'project'
    }

    def __init__(
        self,
        connection: 'Connection',
        id: str | None = None,
        project: Project | str | None = None,
        default_connection_map: bool = False,
        ds_connection: DatasourceConnection | None = None,
        datasource: DatasourceInstance | None = None,
        user: User | UserGroup | None = None,
        login: DatasourceLogin | None = None,
        locale: Language | None = None,
    ):
        """Initialise connection mapping by passing the ID or by passing
        True for `default_connection_map` and the project for which to
        fetch the default mapping.

        To list all connection mappings use the
        `list_datasource_mappings()` method.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            id: ID of connection mapping
            project: Project object or ID for restricting the search to
                just this project.
            default_connection_map: Whether to fetch the project's
                default mapping.
        """
        if id is None:
            if project is None or not default_connection_map:
                raise ValueError(
                    "Please either specify the `id` or `default_connection_map` and"
                    " the `project` to find the default mapping for the given project."
                )

            mappings: list[dict] = self._list(
                connection=connection,
                default_connection_map=True,
                project=project,
                to_dictionary=True,
            )

            if not mappings:
                project_id = get_objects_id(project, Project)

                raise ValueError(
                    f"The project: {project_id} has no default datasource mapping."
                )
            elif len(mappings) > 1:
                project_id = get_objects_id(project, Project)

                raise ValueError(
                    f"The project: {project_id} has more than one default connection "
                    f"mapping."
                )
            else:
                data = mappings[0]
                data.update({'ds_connection': data.pop('connection')})
                self._init_variables(connection=connection, **mappings[0])

        else:
            super().__init__(
                connection,
                id,
                default_connection_map=default_connection_map,
                project_id=project,
            )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        if project_id := kwargs.get('project_id'):
            self.project = Project.from_dict({'id': project_id}, self.connection)
            self._project_id = project_id

        if ds_connection := kwargs.get('ds_connection'):
            self.ds_connection = DatasourceConnection.from_dict(
                ds_connection, self.connection
            )

        if datasource := kwargs.get('datasource'):
            self.datasource = DatasourceInstance.from_dict(datasource, self.connection)

        if user := kwargs.get('user'):
            self.user = User.from_dict(user, self.connection)

            if self.user.subtype == 8705:
                self.user = UserGroup.from_dict(user, self.connection)

        if login := kwargs.get('login'):
            self.login = DatasourceLogin.from_dict(login, self.connection)

        if locale := kwargs.get('locale'):
            self.locale = Language.from_dict(locale, self.connection)
        else:
            self.locale = None

        self.default_connection_map = kwargs.get('default_connection_map')

    @classmethod
    def _list(
        cls,
        connection: 'Connection',
        project: Project | str | None = None,
        to_dictionary: bool = False,
        limit: int | None = None,
        user: User | UserGroup | str | None = None,
        ds_connection: DatasourceConnection | str | None = None,
        datasource: DatasourceInstance | str | None = None,
        login: DatasourceLogin | str | None = None,
        locale: Language | str | None = None,
        default_connection_map: bool = False,
    ) -> list['DatasourceMap'] | list[dict]:
        project_id = get_objects_id(project, Project)

        if isinstance(locale, Language):
            locale = {'id': locale.id, 'name': locale.name}
        elif isinstance(locale, str):
            locale = DatasourceMap._get_locale(connection=connection, query=locale)
            locale = {'id': locale.id, 'name': locale.name}

        filters = delete_none_values(
            delete_none_values(
                {
                    'user': {
                        'id': get_objects_id(user, User)
                        or get_objects_id(user, UserGroup),
                    },
                    'connection': {
                        'id': get_objects_id(ds_connection, DatasourceConnection),
                    },
                    'datasource': {
                        'id': get_objects_id(datasource, DatasourceInstance),
                    },
                    'login': {
                        'id': get_objects_id(login, DatasourceLogin),
                    },
                    'locale': locale,
                },
                recursion=True,
            ),
            recursion=False,
        )

        mappings = datasources_processors.get_mappings(
            connection=connection,
            project_id=project_id,
            limit=limit,
            default_connection_map=default_connection_map,
            filters=filters if filters else None,
        )

        if to_dictionary:
            return mappings
        else:
            return [
                cls.from_dict(source=item, connection=connection) for item in mappings
            ]

    @classmethod
    def create(
        cls,
        connection: 'Connection',
        project: Project | str,
        user: User | UserGroup | str,
        ds_connection: DatasourceConnection | str,
        datasource: DatasourceInstance | str,
        login: DatasourceLogin | str,
        locale: Language | str | None = None,
        locale_id: str | None = None,
        locale_name: str | None = None,
    ) -> 'DatasourceMap':
        """Create a new connection mapping on the server.
        If more than one locale related parameters are provided,
        `locale` has priority, then `locale_id`.

        Args:
            connection: A Strategy One connection object
            project: The project the mapping is to be assigned to
            user: The User or UserGroup to be mapped
            ds_connection: The Datasource Connection to be mapped
            datasource: The Datasource Instance to be mapped
            login: The Datasource Login to be mapped
            locale: The locale to be mapped.
            locale_id: The id of locale to be mapped.
            locale_name: The name of locale to be mapped.

        Returns:
            DatasourceMap object
        """
        project_id = get_objects_id(project, Project)
        user_id = get_objects_id(user, User) or get_objects_id(user, UserGroup)
        connection_id = get_objects_id(ds_connection, DatasourceConnection)
        datasource_id = get_objects_id(datasource, DatasourceInstance)
        login_id = get_objects_id(login, DatasourceLogin)
        body = {
            'projectId': project_id,
            'user': {'id': user_id},
            'connection': {'id': connection_id},
            'datasource': {'id': datasource_id},
            'login': {'id': login_id},
        }
        if locale and isinstance(locale, Language):
            body['locale'] = {'id': locale.id, 'name': locale.name}
        elif locale and isinstance(locale, str):
            locale = DatasourceMap._get_locale(connection=connection, query=locale)
            body['locale'] = {'id': locale.id, 'name': locale.name}
        elif locale_id and isinstance(locale_id, str):
            locale = Language(connection=connection, id=locale_id)
            body['locale'] = {'id': locale.id, 'name': locale.name}
        elif locale_name and isinstance(locale_name, str):
            locale = Language(connection=connection, name=locale_name)
            body['locale'] = {'id': locale.id, 'name': locale.name}

        response_data = datasources.create_datasource_mapping(
            connection=connection, body=body
        ).json()
        if config.verbose:
            logger.info(
                "Successfully created datasource connection mapping with ID:"
                f" '{response_data.get('id')}'"
            )
        return cls.from_dict(
            source=helper.camel_to_snake(response_data), connection=connection
        )

    @staticmethod
    def __translate_names(name: str):
        return 'ds_connection' if name == 'connection' else name

    @classmethod
    def from_dict(
        cls,
        source: dict[str, Any],
        connection: 'Connection',
        to_snake_case: bool = True,
    ) -> 'DatasourceMap':
        data = {cls.__translate_names(key): val for key, val in source.items()}
        return super().from_dict(data, connection, to_snake_case)

    def to_dict(self, camel_case=True):
        data = super().to_dict(camel_case=camel_case)
        return {self.__translate_names(key): val for key, val in data.items()}

    def alter(
        self,
        user: User | UserGroup | str | None = None,
        ds_connection: DatasourceConnection | str | None = None,
        datasource: DatasourceInstance | str | None = None,
        login: DatasourceLogin | str | None = None,
        locale: Language | str | None = None,
    ):
        """Replace the connection mapping with a newly created one
        with field values copied unless new ones are specified.

        Args:
            connection: A Strategy One connection object
            user: The User or UserGroup to be mapped
            ds_connection: The Datasource Connection to be mapped
            datasource: The Datasource Instance to be mapped
            login: The Datasource Login to be mapped
            locale: The locale to be mapped

        Returns:
            DatasourceMap object
        """
        if (
            user is None
            and ds_connection is None
            and datasource is None
            and login is None
            and locale is None
        ):
            if config.verbose:
                logger.info(
                    f"No changes specified for {type(self).__name__} with ID:'"
                    f"{self.id}'."
                )
            return

        self.delete(force=True)

        new_conn_map = self.create(
            connection=self.connection,
            project=self.project,
            user=user or self.user,
            ds_connection=ds_connection or self.ds_connection,
            datasource=datasource or self.datasource,
            login=login or self.login,
            locale=locale or self.locale,
        )
        self._id = new_conn_map.id
        self.fetch()

    @staticmethod
    def _get_locale(connection: Connection, query: str):
        locales_list = list_languages(connection=connection, base_language_lcid=0)
        for locale in locales_list:
            if query in [locale.id, locale.name, locale.abbreviation]:
                return locale
        raise ValueError(
            f'Locale with id or name or abbreviation: {query} does not exists.'
        )
